from __future__ import annotations

import asyncio
import logging

import aiohttp
from typing import Optional, TYPE_CHECKING, Tuple, Type, Dict, Callable, Generic, TypeVar, Awaitable, Union, cast, List
from . import models, http
from .models import _loads
from twitchio import PartialUser, Unauthorized, HTTPException

if TYPE_CHECKING:
    from typing_extensions import Literal
    from twitchio import Client

logger = logging.getLogger("twitchio.ext.eventsub.ws")

_message_types = {
    "notification": models.NotificationEvent,
    "revocation": models.RevokationEvent,
    "session_reconnect": models.ReconnectEvent,
    "session_keepalive": models.KeepAliveEvent,
}
_messages = Union[models.NotificationEvent, models.RevokationEvent, models.ReconnectEvent, models.KeepAliveEvent]


class _Subscription:
    __slots__ = "event", "condition", "token", "subscription_id", "cost", "created"

    def __init__(self, event_type: Tuple[str, int, Type[models.EventData]], condition: Dict[str, str], token: str):
        self.event = event_type
        self.condition = condition
        self.token = token
        self.subscription_id: Optional[str] = None
        self.cost: Optional[int] = None
        self.created: asyncio.Future[Tuple[Literal[False], int] | Tuple[Literal[True], None]] | None = asyncio.Future()


_T = TypeVar("_T")


class _WakeupList(list, Generic[_T]):
    def __init__(self, *args):
        super().__init__(*args)
        self._append_waiters = []
        self._pop_waiters = []

    def _wakeup_append(self, obj: _T) -> None:
        try:
            loop = asyncio.get_running_loop()
            for cb in self._append_waiters:
                loop.create_task(cb(obj))
        except:  # don't wake the waiters if theres no loop
            pass

    def _wakeup_pop(self, obj: _T) -> None:
        try:
            loop = asyncio.get_running_loop()
            for cb in self._pop_waiters:
                loop.create_task(cb(obj))
        except:  # don't wake the waiters
            pass

    def append(self, obj: _T) -> None:
        self._wakeup_append(obj)
        super().append(obj)

    def insert(self, index: int, obj: _T) -> None:
        self._wakeup_append(obj)
        super().insert(index, obj)

    def __delitem__(self, key: int):
        self._wakeup_pop(self[key])
        super().__delitem__(key)

    def pop(self, index: int = ...) -> _T:
        resp = super().pop(index)
        self._wakeup_pop(resp)
        return resp

    def add_append_callback(self, cb: Callable[[_T], Awaitable[None]]) -> None:
        self._append_waiters.append(cb)

    def add_pop_callback(self, cb: Callable[[_T], Awaitable[None]]) -> None:
        self._pop_waiters.append(cb)


class Websocket:
    URL = "wss://eventsub.wss.twitch.tv/ws"

    def __init__(self, client: Client, http: http.EventSubHTTP):
        self.client = client
        self._http = http
        self._subscription_pool = _WakeupList[_Subscription]()
        self._subscription_pool.add_append_callback(self._wakeup_and_connect)
        self._sock: Optional[aiohttp.ClientWebSocketResponse] = None
        self._pump_task: Optional[asyncio.Task] = None
        self._timeout: Optional[int] = None
        self._session_id: Optional[str] = None
        self._target_user_id: int | None = None  # each websocket can only have one authenticated user on it for some bizzare reason, but this isnt documented anywhere
        self.remaining_slots: int = 300  # default to 300

    def __hash__(self) -> int:
        return hash(self.session_id)

    def __eq__(self, __value: object) -> bool:
        return __value is self

    @property
    def session_id(self) -> Optional[str]:
        return self._session_id

    @property
    def is_connected(self) -> bool:
        return self._sock is not None and not self._sock.closed

    async def _subscribe(self, obj: _Subscription) -> dict | None:
        try:
            resp = await self._http.create_websocket_subscription(obj.event, obj.condition, self._session_id, obj.token)
        except HTTPException as e:
            if obj.created:
                obj.created.set_result((False, e.status))

            else:
                logger.error(
                    "An error (%s %s) occurred while attempting to resubscribe to an event on reconnect: %s",
                    e.status,
                    e.reason,
                    e.message,
                )

            return None

        if obj.created:
            obj.created.set_result((True, None))

        data = resp["data"][0]
        self.remaining_slots = resp["max_total_cost"] - resp["total_cost"]
        obj.cost = data["cost"]

        return data

    def add_subscription(self, sub: _Subscription) -> None:
        self._subscription_pool.append(sub)

    async def _wakeup_and_connect(self, obj: _Subscription):
        if self.is_connected:
            await self._subscribe(obj)
            return

    async def connect(self, reconnect_url: Optional[str] = None):
        async with aiohttp.ClientSession() as session:
            sock = self._sock = await session.ws_connect(reconnect_url or self.URL)
            session.detach()

        welcome = await sock.receive_json(loads=_loads, timeout=3)
        logger.debug("Received websocket payload: %s", welcome)
        self._session_id = welcome["payload"]["session"]["id"]
        self._timeout = welcome["payload"]["session"]["keepalive_timeout_seconds"]

        logger.debug("Created websocket connection with session ID: %s and timeout %s", self._session_id, self._timeout)

        self._pump_task = self.client.loop.create_task(self.pump())

        if reconnect_url:  # don't resubscribe to events
            return

        for sub in self._subscription_pool:
            await self._subscribe(sub)

    async def pump(self) -> None:
        sock: aiohttp.ClientWebSocketResponse = cast(aiohttp.ClientWebSocketResponse, self._sock)
        while self.is_connected:
            try:
                msg = await sock.receive_str(
                    timeout=self._timeout + 1
                )  # extra jitter on the timeout in case of network lag
                if not msg:
                    logger.warning("Received empty payload ")

                logger.debug("Received websocket payload: %s", msg)
                frame: _messages = self.parse_frame(_loads(msg))
                self.client.run_event("eventsub_debug", frame)

                if isinstance(frame, models.NotificationEvent):
                    self.client.run_event(
                        f"eventsub_notification_{models.SubscriptionTypes._name_map[frame.subscription.type]}", frame
                    )
                    self.client.run_event("eventsub_notification", frame)

                elif isinstance(frame, models.RevokationEvent):
                    self.client.run_event("eventsub_revokation", frame)

                elif isinstance(frame, models.KeepAliveEvent):
                    self.client.run_event("eventsub_keepalive", frame)

                elif isinstance(frame, models.ReconnectEvent):
                    self.client.run_event("eventsub_reconnect", frame)
                    self._sock = None
                    await self.connect(frame.reconnect_url)
                    await sock.close(code=aiohttp.WSCloseCode.GOING_AWAY, message=b"reconnecting")
                    return

            except asyncio.TimeoutError:
                logger.warning(f"Websocket timed out (timeout: {self._timeout}), reconnecting")
                await cast(aiohttp.ClientWebSocketResponse, self._sock).close(
                    code=aiohttp.WSCloseCode.ABNORMAL_CLOSURE, message=b"timeout surpassed"
                )
                await self.connect()
                return

            except TypeError as e:
                logger.warning(f"Received bad frame: {e.args[0]}")

                if "257" in e.args[0]:  # websocket was closed, reconnect
                    logger.info("Known bad frame, restarting connection")
                    await self.connect()
                    return

            except Exception as e:
                logger.error("Exception in the pump function!", exc_info=e)
                raise

    def parse_frame(self, frame: dict) -> _messages:
        type_: str = frame["metadata"]["message_type"]
        return _message_types[type_](self, frame, None)


class EventSubWSClient:
    def __init__(self, client: Client):
        self.client = client
        self._http: http.EventSubHTTP = http.EventSubHTTP(self, token=None)

        self._sockets: List[Websocket] = []
        self._ready_to_subscribe: List[_Subscription] = []

    async def _assign_subscription(self, sub: _Subscription) -> None:
        if not self._sockets:
            w = Websocket(self.client, self._http)
            await w.connect()

            self._sockets.append(w)

        success = False
        bad_sockets: set[Websocket] | None = None  # dont allocate unless we need it

        while not success:
            s: Websocket | None = None  # really it'll never be none after this point, but ok pyright

            if bad_sockets is not None:
                socks = filter(lambda sock: sock not in bad_sockets, self._sockets)  # type: ignore
            else:
                socks = self._sockets

            for s in socks:
                if s.remaining_slots > 0:
                    s.add_subscription(sub)
                    break

            else:  # there are no sockets, create one and break
                s = Websocket(self.client, self._http)
                await s.connect()

                s.add_subscription(sub)
                return

            assert sub.created is not None  # go away pyright

            success, status = await sub.created

            if not success and status == 400:
                # can't be on that socket due to someone else being on it, try again on a different one
                if bad_sockets is None:
                    bad_sockets = set()

                bad_sockets.add(s)
                sub.created = asyncio.Future()
                continue

            elif not success and status in (401, 403):
                raise Unauthorized("You are not authorized to make this subscription", status=status)

            elif not success:
                raise RuntimeError(f"Subscription failed, reason unknown. Status: {status}")

            else:
                sub.created = None  # don't need that future to sit in memory
                break

    async def subscribe_user_updated(self, user: Union[PartialUser, str, int], token: str):
        if isinstance(user, PartialUser):
            user = user.id

        user = str(user)
        sub = _Subscription(models.SubscriptionTypes.user_update, {"user_id": user}, token)
        await self._assign_subscription(sub)

    async def subscribe_channel_raid(
        self,
        token: str,
        from_broadcaster: Union[PartialUser, str, int] = None,
        to_broadcaster: Union[PartialUser, str, int] = None,
    ):
        if (not from_broadcaster and not to_broadcaster) or (from_broadcaster and to_broadcaster):
            raise ValueError("Expected 1 of from_broadcaster or to_broadcaster")

        if from_broadcaster:
            who = "from_broadcaster_user_id"
            broadcaster = from_broadcaster
        else:
            who = "to_broadcaster_user_id"
            broadcaster = to_broadcaster

        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id

        broadcaster = str(broadcaster)
        sub = _Subscription(models.SubscriptionTypes.raid, {who: broadcaster}, token)
        await self._assign_subscription(sub)

    async def _subscribe_channel_points_reward(
        self,
        event: Tuple[str, int, Type[models._DataType]],
        broadcaster: Union[PartialUser, str, int],
        token: str,
        reward_id: str = None,
    ):
        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id

        broadcaster = str(broadcaster)
        data = {"broadcaster_user_id": broadcaster}
        if reward_id:
            data["reward_id"] = reward_id

        sub = _Subscription(event, data, token)
        await self._assign_subscription(sub)

    async def _subscribe_with_broadcaster(
        self, event: Tuple[str, int, Type[models._DataType]], broadcaster: Union[PartialUser, str, int], token: str
    ):
        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id

        broadcaster = str(broadcaster)
        sub = _Subscription(event, {"broadcaster_user_id": broadcaster}, token)
        await self._assign_subscription(sub)

    async def _subscribe_with_broadcaster_moderator(
        self,
        event: Tuple[str, int, Type[models._DataType]],
        broadcaster: Union[PartialUser, str, int],
        moderator: Union[PartialUser, str, int],
        token: str,
    ):
        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id
        if isinstance(moderator, PartialUser):
            moderator = moderator.id

        broadcaster = str(broadcaster)
        moderator = str(moderator)
        sub = _Subscription(event, {"broadcaster_user_id": broadcaster, "moderator_user_id": moderator}, token)
        await self._assign_subscription(sub)

    async def subscribe_channel_bans(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.ban, broadcaster, token)

    async def subscribe_channel_unbans(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.unban, broadcaster, token)

    async def subscribe_channel_subscriptions(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription, broadcaster, token)

    async def subscribe_channel_subscription_end(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription_end, broadcaster, token)

    async def subscribe_channel_subscription_gifts(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription_gift, broadcaster, token)

    async def subscribe_channel_subscription_messages(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription_message, broadcaster, token)

    async def subscribe_channel_cheers(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.cheer, broadcaster, token)

    async def subscribe_channel_update(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_update, broadcaster, token)

    async def subscribe_channel_follows(self, broadcaster: Union[PartialUser, str, int], token: str):
        raise RuntimeError("This subscription has been removed by twitch, please use subscribe_channel_follows_v2")

    async def subscribe_channel_follows_v2(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.followV2, broadcaster, moderator, token
        )

    async def subscribe_channel_moderators_add(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_moderator_add, broadcaster, token)

    async def subscribe_channel_moderators_remove(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_moderator_remove, broadcaster, token)

    async def subscribe_channel_goal_begin(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_goal_begin, broadcaster, token)

    async def subscribe_channel_goal_progress(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_goal_progress, broadcaster, token)

    async def subscribe_channel_goal_end(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_goal_end, broadcaster, token)

    async def subscribe_channel_hypetrain_begin(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.hypetrain_begin, broadcaster, token)

    async def subscribe_channel_hypetrain_progress(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.hypetrain_progress, broadcaster, token)

    async def subscribe_channel_hypetrain_end(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.hypetrain_end, broadcaster, token)

    async def subscribe_channel_stream_start(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.stream_start, broadcaster, token)

    async def subscribe_channel_stream_end(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.stream_end, broadcaster, token)

    async def subscribe_channel_points_reward_added(
        self, broadcaster: Union[PartialUser, str, int], reward_id: str, token: str
    ):
        await self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_add, broadcaster, token, reward_id
        )

    async def subscribe_channel_points_reward_updated(
        self, broadcaster: Union[PartialUser, str, int], reward_id: str, token: str
    ):
        await self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_update, broadcaster, token, reward_id
        )

    async def subscribe_channel_points_reward_removed(
        self, broadcaster: Union[PartialUser, str, int], reward_id: str, token: str
    ):
        await self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_remove, broadcaster, token, reward_id
        )

    async def subscribe_channel_points_redeemed(
        self, broadcaster: Union[PartialUser, str, int], token: str, reward_id: str = None
    ):
        await self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_redeem, broadcaster, token, reward_id
        )

    async def subscribe_channel_points_redeem_updated(
        self, broadcaster: Union[PartialUser, str, int], token: str, reward_id: str = None
    ):
        await self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_redeem_updated, broadcaster, token, reward_id
        )

    async def subscribe_channel_poll_begin(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.poll_begin, broadcaster, token)

    async def subscribe_channel_poll_progress(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.poll_progress, broadcaster, token)

    async def subscribe_channel_poll_end(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.poll_end, broadcaster, token)

    async def subscribe_channel_prediction_begin(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_begin, broadcaster, token)

    async def subscribe_channel_prediction_progress(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_progress, broadcaster, token)

    async def subscribe_channel_prediction_lock(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_lock, broadcaster, token)

    async def subscribe_channel_prediction_end(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_end, broadcaster, token)

    async def subscribe_channel_auto_reward_redeem(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.auto_reward_redeem, broadcaster, token)

    async def subscribe_channel_shield_mode_begin(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shield_mode_begin, broadcaster, moderator, token
        )

    async def subscribe_channel_shield_mode_end(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shield_mode_end, broadcaster, moderator, token
        )

    async def subscribe_channel_shoutout_create(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shoutout_create, broadcaster, moderator, token
        )

    async def subscribe_channel_shoutout_receive(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shoutout_receive, broadcaster, moderator, token
        )

    async def subscribe_channel_charity_donate(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_charity_donate, broadcaster, token)

    async def subscribe_channel_unban_request_create(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.unban_request_create, broadcaster, moderator, token
        )

    async def subscribe_channel_unban_request_resolve(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.unban_request_resolve, broadcaster, moderator, token
        )

    async def subscribe_automod_message_hold(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_message_hold, broadcaster, moderator, token
        )

    async def subscribe_automod_message_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_message_update, broadcaster, moderator, token
        )

    async def subscribe_automod_settings_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_settings_update, broadcaster, moderator, token
        )

    async def subscribe_automod_terms_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_terms_update, broadcaster, moderator, token
        )

    async def subscribe_suspicious_user_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.suspicious_user_update, broadcaster, moderator, token
        )

    async def subscribe_channel_moderate(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int], token: str
    ):
        await self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_moderate, broadcaster, moderator, token
        )

    async def subscribe_channel_vip_add(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_vip_add, broadcaster, token)

    async def subscribe_channel_vip_remove(self, broadcaster: Union[PartialUser, str, int], token: str):
        await self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_vip_remove, broadcaster, token)



from __future__ import annotations
from typing import TYPE_CHECKING, Tuple, Dict, Type, Union, Optional
from ...http import Route

from . import models

if TYPE_CHECKING:
    from .server import EventSubClient
    from .websocket import EventSubWSClient
    from .models import EventData, Subscription

__all__ = ("EventSubHTTP",)


class EventSubHTTP:
    def __init__(self, client: Union[EventSubClient, EventSubWSClient], token: Optional[str]):
        self._client = client
        self._http = client.client._http
        self._token = token

    async def create_webhook_subscription(
        self, event_type: Tuple[str, int, Type[EventData]], condition: Dict[str, str]
    ):
        payload = {
            "type": event_type[0],
            "version": str(event_type[1]),
            "condition": condition,
            "transport": {"method": "webhook", "callback": self._client.route, "secret": self._client.secret},
        }
        route = Route("POST", "eventsub/subscriptions", body=payload, token=self._token)
        return await self._http.request(route, paginate=False, force_app_token=True)

    async def create_websocket_subscription(
        self, event_type: Tuple[str, int, Type[EventData]], condition: Dict[str, str], session_id: str, token: str
    ) -> dict:
        payload = {
            "type": event_type[0],
            "version": str(event_type[1]),
            "condition": condition,
            "transport": {"method": "websocket", "session_id": session_id},
        }
        route = Route("POST", "eventsub/subscriptions", body=payload, token=token)
        return await self._http.request(route, paginate=False, full_body=True)  # type: ignore

    async def delete_subscription(self, subscription: Union[str, Subscription]):
        if isinstance(subscription, models.Subscription):
            return await self._http.request(
                Route("DELETE", "eventsub/subscriptions", query=[("id", subscription.id)]), paginate=False
            )
        return await self._http.request(
            Route("DELETE", "eventsub/subscriptions", query=[("id", subscription)]), paginate=False
        )

    async def get_subscriptions(
        self, status: Optional[str] = None, sub_type: Optional[str] = None, user_id: Optional[int] = None
    ):
        qs = []
        if status:
            qs.append(("status", status))
        if sub_type:
            qs.append(("type", sub_type))
        if user_id:
            qs.append(("user_id", str(user_id)))
        if len(qs) > 1:
            raise ValueError("You cannot specify more than one filter.")

        return [
            models.Subscription(d)
            for d in await self._http.request(Route("GET", "eventsub/subscriptions", query=qs), paginate=False)
        ]

    async def get_status(self, status: str = None):
        qs = []
        if status:
            qs.append(("status", status))

        v = await self._http.request(Route("GET", "eventsub/subscriptions", query=qs), paginate=False, full_body=True)
        del v["data"]
        del v["pagination"]
        return v



"""MIT License

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import asyncio
import audioop
import dataclasses
import io
import logging
import subprocess
import threading
import time
from functools import partial
from typing import Any, Callable, Coroutine, Dict, List, Optional, TypeVar, Union

import pyaudio
from yt_dlp import YoutubeDL
from tinytag import TinyTag


__all__ = ("Sound", "AudioPlayer")


logger = logging.getLogger(__name__)


AP = TypeVar("AP", bound="AudioPlayer")

ffmpeg_bin: Optional[str] = None
try:
    proc = subprocess.Popen(["ffmpeg", "-version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
except FileNotFoundError:
    try:
        proc = subprocess.Popen(["ffmpeg.exe", "-version"])
    except FileNotFoundError:
        ffmpeg_bin = None
    else:
        ffmpeg_bin = "ffmpeg.exe"
else:
    ffmpeg_bin = "ffmpeg"


__all__ = ("Sound", "AudioPlayer")


@dataclasses.dataclass
class OutputDevice:
    """Class which represents an OutputDevice usable with :class:`AudioPlayer` .

    Pass this into the appropriate :class:`AudioPlayer` method to set or change device.

    Attributes
    ----------
    name: :class:`str`
        The name of the device.
    index: :class:`int`
        The index of the device.
    channels: :class:`int`
        The amount of available channels this device has.
    """

    name: str
    index: int
    channels: int


class Sound:
    """TwitchIO Sound Source.

    This is the class that represents a TwitchIO Sound, able to be played in :class:`AudioPlayer`.


    .. warning::

        This class is still in Beta stages and currently only supports generating sounds via :meth:`ytdl_search`
        or locally downloaded audio files. Future versions will have support for file like objects.

    Parameters
    __________
    source: Optional[Union[str, io.BufferedIOBase]]
        The source to play. Could be a string representing a local file or bytes.
        To search YouTube, use :meth:`ytdl_search` instead.
    info: Optional[dict]
        The info dict provided via ytdl. Only available when searching via YouTube. Do not pass this directly.

    Attributes
    ----------
    title: str
        The title of the track. Will be None if :class:`io.BufferedIOBase` is supplied as a source.
    url: Optional[str]
        The url of the track searched via YouTube. Will be None if source is supplied.
    """

    CODECS = (
        "mp3",
        "aac",
        "flv",
        "wav",
        "opus",
    )

    YTDLOPTS = {
        "format": "bestaudio/best",
        "restrictfilenames": True,
        "noplaylist": True,
        "nocheckcertificate": True,
        "ignoreerrors": False,
        "logtostderr": False,
        "quiet": True,
        "progress": False,
        "no_warnings": True,
        "default_search": "auto",
        "source_address": "0.0.0.0",  # ipv6 addresses cause issues sometimes
    }

    YTDL = YoutubeDL(YTDLOPTS)

    def __init__(
        self,
        source: Optional[Union[str, io.BufferedIOBase]] = None,
        *,
        info: Optional[dict] = None,
    ):
        self.title = None
        self.url = None

        self.proc = None

        if ffmpeg_bin is None:
            raise RuntimeError("ffmpeg is required to create and play Sounds. Check your is present in your Path")

        if info:
            self.proc = subprocess.Popen(
                [
                    ffmpeg_bin,
                    "-reconnect",
                    "1",
                    "-reconnect_streamed",
                    "1",
                    "-reconnect_delay_max",
                    "5",
                    "-i",
                    info["url"],
                    "-loglevel",
                    "panic",
                    "-vn",
                    "-f",
                    "s16le",
                    "pipe:1",
                ],
                stdout=subprocess.PIPE,
            )

            self.title = info.get("title", None)
            self.url = info.get("url", None)

        elif isinstance(source, str):
            self.title = source
            tag = TinyTag.get(source)
            self._rate = tag.samplerate
            self._channels = tag.channels

            self.proc = subprocess.Popen(
                [
                    ffmpeg_bin,
                    "-i",
                    source,
                    "-loglevel",
                    "panic",
                    "-vn",
                    "-f",
                    "s16le",
                    "pipe:1",
                ],
                stdout=subprocess.PIPE,
            )

    @classmethod
    async def ytdl_search(cls, search: str, *, loop: Optional[asyncio.BaseEventLoop] = None):
        """|coro|

        Search songs via YouTube ready to be played.
        """
        loop = loop or asyncio.get_event_loop()

        to_run = partial(cls.YTDL.extract_info, url=search, download=False)
        data = await loop.run_in_executor(None, to_run)

        if "entries" in data:
            # take first item from a playlist
            data = data["entries"][0]

        data["time"] = time.time()

        return cls(info=data)

    @property
    def channels(self):
        """The audio source channels."""
        return self._channels

    @channels.setter
    def channels(self, channels: int):
        """Set audio source channels."""
        self._channels = channels

    @property
    def rate(self):
        """The audio source sample rate."""
        return self._rate

    @rate.setter
    def rate(self, rate: int):
        """Set audio source sample rate."""
        self._rate = rate

    @property
    def source(self):
        """The raw audio source."""
        return self._source

    def __del__(self):
        self._source = None

        if self.proc:
            self.proc.kill()


class AudioPlayer:
    """TwitchIO Sounds Audio Player.

    Use this class to control and play sounds generated with :class:`Sound` .

    Parameters
    ----------
    callback: Coroutine
        The coroutine called when a Sound has finished playing.


    .. note ::

        To use this player with the system non default output device, see :meth:`with_device`, or
        :attr:`active_device`. The output device can be changed live by setting `active_device` with an
        :class:`OutputDevice`.
    """

    def __init__(self, *, callback: Callable[..., Coroutine[Any, Any, None]]):
        self._pa = pyaudio.PyAudio()
        self._stream = None

        self._current: Sound = None  # type: ignore

        self._paused: bool = False
        self._playing: bool = False
        self._volume = 100

        self._callback = callback
        self._thread: threading.Thread = None  # type: ignore

        self._loop = asyncio.get_event_loop()

        self._devices: Dict[int, OutputDevice] = {}
        self._get_devices()

        self._active_devices: List[OutputDevice] = []
        self._use_device: Optional[OutputDevice] = None

    def _get_devices(self):
        for index in range(self._pa.get_device_count()):
            device = self._pa.get_device_info_by_index(index)

            if device["maxInputChannels"] != 0:
                continue

            self._devices[index] = OutputDevice(
                name=device["name"],
                index=device["index"],
                channels=device["maxOutputChannels"],
            )

    def play(self, sound: Sound, *, replace: bool = False) -> None:
        """Play a :class:`Sound` object.

        When play has finished playing, the event :meth:`event_player_finished` is called.
        """
        if not isinstance(sound, Sound):
            raise TypeError("sound parameter must be of type <Sound>.")

        if self._playing and replace:
            self.stop()
        elif self._playing and not replace:
            return

        self._thread = threading.Thread(target=self._play_run, args=([sound]), daemon=True)
        self._thread.start()

    def _play_run(self, sound: Sound):
        self._current = sound

        device = self._use_device.index if self._use_device else None
        self._stream = self._pa.open(
            format=pyaudio.paInt16,
            output=True,
            channels=sound.channels,
            rate=sound.rate,
            output_device_index=device,
        )

        bytes_ = sound.proc.stdout.read(4096)

        self._playing = True
        while self._playing:
            while self._paused:
                time.sleep(0.1)

            if self._playing is False:
                break

            if not bytes_:
                break

            self._stream.write(audioop.mul(bytes_, 2, self._volume / 100))
            bytes_ = sound.proc.stdout.read(4096)

        self._clean(sound)
        asyncio.run_coroutine_threadsafe(self._callback(), loop=self._loop)

    def _clean(self, sound: Sound):
        sound.proc.kill()
        self._stream = None

        self._current = None
        self._paused = False
        self._playing = False

    @property
    def source(self):
        """The currently playing :class:`Sound` object."""
        return self._current

    @property
    def current(self):
        """An alias to :meth:`source`."""
        return self.source

    def pause(self):
        """Method which pauses the player."""
        self._paused = True

    def resume(self):
        """Method which resumes the player."""
        self._paused = False

    @property
    def is_paused(self) -> bool:
        """Property which returns whether the player is currently paused."""
        return self._paused

    def stop(self):
        """Stops the player and clears the source."""
        self._playing = False

    @property
    def is_playing(self):
        """Property which returns whether the player is currently playing."""
        return self._playing

    @property
    def volume(self) -> int:
        """Property which returns the current volume.

        This property can also be set with a volume between 1 and 100, to change the volume.
        """
        return self._volume

    @volume.setter
    def volume(self, level: int) -> None:
        if level > 100 or level < 1:
            raise ValueError("Volume must be between 1 and 100")

        self._volume = level

    @property
    def devices(self) -> Dict[int, OutputDevice]:
        """Return a dict of :class:`OutputDevice` that can be used to output audio."""
        return self._devices

    @property
    def active_device(self) -> Optional[OutputDevice]:
        """Return the active output device for this player. Could be None if default is being used.

        This property can also be set with a new :class:`OutputDevice` to change audio output.
        """
        return self._use_device

    @active_device.setter
    def active_device(self, device: OutputDevice) -> None:
        if not isinstance(device, OutputDevice):
            raise TypeError(f"Parameter <device> must be of type <{type(OutputDevice)}> not <{type(device)}>.")

        self._use_device = device
        if not self._stream:
            return

        if self._playing:
            self._paused = True

        self._stream = self._pa.open(
            format=pyaudio.paInt16,
            output=True,
            channels=self._current.channels,
            rate=self._current.rate,
            output_device_index=device.index,
        )

        self._paused = False

    @classmethod
    def with_device(cls, *, callback: Callable[..., Coroutine[Any, Any, None]], device: OutputDevice) -> AP:
        """Method which returns a player ready to be used with the given :class:`OutputDevice`.

        Returns
        -------
            :class:`OutputDevice`
        """
        if not isinstance(device, OutputDevice):
            raise TypeError(f"Parameter <device> must be of type <{type(OutputDevice)}> not <{type(device)}>.")

        self = cls(callback=callback)
        self._use_device = device

        return self



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import datetime
from typing import Optional, TYPE_CHECKING

from .errors import HTTPException, Unauthorized
from .utils import parse_timestamp

if TYPE_CHECKING:
    from .http import TwitchHTTP
    from .user import PartialUser
__all__ = "CustomReward", "CustomRewardRedemption"


class CustomReward:
    """
    Represents a Custom Reward object, as given by the api. Use :func:`~twitchio.PartialUser.get_custom_rewards` to fetch these

    Attributes
    -----------
    id: :class:`str`
        The id of the custom reward
    title: :class:`str`
        The title of the custom reward
    image: :class:`str`
        The url of the image of the custom reward
    background_color: :class:`str`
        The background color of the custom reward
    enabled: :class:`bool`
        Whether the custom reward is enabled
    cost: :class:`int`
        The cost of the custom reward
    prompt: :class:`str`
        The prompt of the custom reward
    input_required: :class:`bool`
        Whether the custom reward requires input
    max_per_stream: Tuple[:class:`bool`, :class:`int`]
        Whether the custom reward is limited to a certain amount per stream, and how many
    max_per_user_stream: Tuple[:class:`bool`, :class:`int`]
        Whether the custom reward is limited to a certain amount per user per stream, and how many
    cooldown: Tuple[:class:`bool`, :class:`int`]
        Whether the custom reward has a cooldown, and how long it is
    paused: :class:`bool`
        Whether the custom reward is paused
    in_stock: :class:`bool`
        Whether the custom reward is in stock
    redemptions_skip_queue: :class:`bool`
        Whether the custom reward's redemptions skip the request queue
    redemptions_current_stream: :class:`bool`
        Whether the custom reward's redemptions are only valid for the current stream
    cooldown_until: :class:`datetime.datetime`
        The datetime the custom reward's cooldown will expire
    """

    __slots__ = (
        "_http",
        "_channel",
        "id",
        "image",
        "background_color",
        "enabled",
        "cost",
        "title",
        "prompt",
        "input_required",
        "max_per_stream",
        "max_per_user_stream",
        "cooldown",
        "paused",
        "in_stock",
        "redemptions_skip_queue",
        "redemptions_current_stream",
        "cooldown_until",
        "_broadcaster_id",
    )

    def __init__(self, http: "TwitchHTTP", obj: dict, channel: "PartialUser"):
        self._http = http
        self._channel = channel

        try:
            self._broadcaster_id = obj["broadcaster_id"]
        except KeyError:
            self._broadcaster_id = obj["channel_id"]
        self.id = obj["id"]
        self.image = obj["image"]["url_1x"] if obj["image"] else obj["default_image"]["url_1x"]
        self.background_color = obj["background_color"]
        self.enabled = obj["is_enabled"]
        self.cost = obj["cost"]
        self.title = obj["title"]
        self.prompt = obj["prompt"]
        self.input_required = obj["is_user_input_required"]

        try:
            self.max_per_stream = (
                obj["max_per_stream_setting"]["is_enabled"],
                obj["max_per_stream_setting"]["max_per_stream"],
            )
            self.max_per_user_stream = (
                obj["max_per_user_per_stream_setting"]["is_enabled"],
                obj["max_per_user_per_stream_setting"]["max_per_user_per_stream"],
            )
            self.cooldown = (
                obj["global_cooldown_setting"]["is_enabled"],
                obj["global_cooldown_setting"]["global_cooldown_seconds"],
            )
        except KeyError:
            self.max_per_stream = (obj["max_per_stream"]["is_enabled"], obj["max_per_stream"]["max_per_stream"])
            self.max_per_user_stream = (
                obj["max_per_user_per_stream"]["is_enabled"],
                obj["max_per_user_per_stream"]["max_per_user_per_stream"],
            )
            self.cooldown = (
                obj["global_cooldown"]["is_enabled"],
                obj["global_cooldown"]["global_cooldown_seconds"],
            )
        self.paused = obj["is_paused"]
        self.in_stock = obj["is_in_stock"]
        self.redemptions_skip_queue = obj["should_redemptions_skip_request_queue"]
        self.redemptions_current_stream = obj["redemptions_redeemed_current_stream"]
        self.cooldown_until = obj["cooldown_expires_at"]

    async def edit(
        self,
        token: str,
        title: Optional[str] = None,
        prompt: Optional[str] = None,
        cost: Optional[int] = None,
        background_color: Optional[str] = None,
        enabled: Optional[bool] = None,
        input_required: Optional[bool] = None,
        max_per_stream_enabled: Optional[bool] = None,
        max_per_stream: Optional[int] = None,
        max_per_user_per_stream_enabled: Optional[bool] = None,
        max_per_user_per_stream: Optional[int] = None,
        global_cooldown_enabled: Optional[bool] = None,
        global_cooldown: Optional[int] = None,
        paused: Optional[bool] = None,
        redemptions_skip_queue: Optional[bool] = None,
    ):
        """
        Edits the reward. Note that apps can only modify rewards they have made.

        Parameters
        -----------
        token: :class:`str`
            The bearer token for the channel of the reward
        title: Optional[:class:`str`]
            The new title of the reward
        prompt: Optional[:class:`str`]
            The new prompt for the reward
        cost: Optional[:class:`int`]
            The new cost for the reward
        background_color: Optional[:class:`str`]
            The new background color for the reward
        enabled: Optional[:class:`bool`]
            Whether the reward is enabled or not
        input_required: Optional[:class:`bool`]
            Whether user input is required or not
        max_per_stream_enabled: Optional[:class:`bool`]
            Whether the stream limit should be enabled
        max_per_stream: Optional[:class:`int`]
            How many times this can be redeemed per stream
        max_per_user_per_stream_enabled: Optional[:class:`bool`]
            Whether the user stream limit should be enabled
        max_per_user_per_stream: Optional[:class:`int`]
            How many times a user can redeem this reward per stream
        global_cooldown_enabled: Optional[:class:`bool`]
            Whether the global cooldown should be enabled
        global_cooldown: Optional[:class:`int`]
            How many seconds the global cooldown should be
        paused: Optional[:class:`bool`]
            Whether redemptions on this reward should be paused or not
        redemptions_skip_queue: Optional[:class:`bool`]
            Whether redemptions skip the request queue or not

        Returns
        --------
        :class:`CustomReward` itself.
        """

        try:
            data = await self._http.update_reward(
                token,
                self._broadcaster_id,
                self.id,
                title,
                prompt,
                cost,
                background_color,
                enabled,
                input_required,
                max_per_stream_enabled,
                max_per_stream,
                max_per_user_per_stream_enabled,
                max_per_user_per_stream,
                global_cooldown_enabled,
                global_cooldown,
                paused,
                redemptions_skip_queue,
            )
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException(
                    "The custom reward was created by a different application, or channel points are "
                    "not available for the broadcaster (403)",
                    error.args[1],
                    403,
                ) from error
            raise
        else:
            for reward in data:
                if reward["id"] == self.id:
                    self.__init__(self._http, reward, self._channel)
                    break
        return self

    async def delete(self, token: str):
        """
        Deletes the custom reward

        Parameters
        ----------
        token:
            :class:`str` the oauth token of the target channel

        Returns
        --------
        None
        """
        try:
            await self._http.delete_custom_reward(token, self._broadcaster_id, self.id)
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException(
                    "The custom reward was created by a different application, or channel points are "
                    "not available for the broadcaster (403)",
                    error.args[1],
                    403,
                ) from error
            raise

    async def get_redemptions(self, token: str, status: str, sort: str = "OLDEST", first: int = 20):
        """
        Gets redemptions for this reward

        Parameters
        -----------
        token:
            :class:`str` the oauth token of the target channel
        status:
            :class:`str` one of UNFULFILLED, FULFILLED or CANCELED
        sort:
            Optional[:class:`str`] the order redemptions are returned in. One of OLDEST, NEWEST. Default: OLDEST.
        first:
            :class:`int` Number of results to be returned when getting the paginated Custom Reward Redemption objects for a reward.
            Limit: 50. Default: 20.
        """
        try:
            data = await self._http.get_reward_redemptions(
                token, self._broadcaster_id, self.id, status=status, sort=sort, first=first
            )
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException(
                    "The custom reward was created by a different application, or channel points are "
                    "not available for the broadcaster (403)",
                    error.args[1],
                    403,
                ) from error
            raise
        else:
            return [CustomRewardRedemption(x, self._http, self) for x in data]

    def __repr__(self):
        return f"<CustomReward id={self.id} title={self.title} cost={self.cost}>"


class CustomRewardRedemption:
    __slots__ = "_http", "_broadcaster_id", "id", "user_id", "user_name", "input", "status", "redeemed_at", "reward"

    def __init__(self, obj: dict, http: "TwitchHTTP", parent: Optional[CustomReward]):
        self._http = http
        self._broadcaster_id = obj["broadcaster_id"]
        self.id = obj["id"]
        self.user_id = int(obj["user_id"])
        self.user_name = obj["user_name"]
        self.input = obj["user_input"]
        self.status = obj["status"]
        self.redeemed_at = parse_timestamp(obj["redeemed_at"])
        self.reward = parent or obj["reward"]

    def __repr__(self):
        return f"<CustomRewardRedemption id={self.id} user_id={self.user_id} user_name={self.user_name} input={self.input} status={self.status} redeemed_at={self.redeemed_at}>"

    async def fulfill(self, token: str):
        """
        marks the redemption as fulfilled

        Parameters
        ----------
        token:
            :class:`str` the token of the target channel

        Returns
        --------
        itself.
        """
        reward_id = self.reward.id if isinstance(self.reward, CustomReward) else self.reward["id"]
        try:
            data = await self._http.update_reward_redemption_status(
                token, self._broadcaster_id, self.id, reward_id, True
            )
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException(
                    "The custom reward was created by a different application, or channel points are "
                    "not available for the broadcaster (403)",
                    error.args[1],
                    403,
                ) from error
            raise
        else:
            self.__init__(data[0], self._http, self.reward if isinstance(self.reward, CustomReward) else None)
            return self

    async def refund(self, token: str):
        """
        marks the redemption as cancelled

        Parameters
        ----------
        token:
            :class:`str` the token of the target channel

        Returns
        --------
        itself.
        """
        reward_id = self.reward.id if isinstance(self.reward, CustomReward) else self.reward["id"]
        try:
            data = await self._http.update_reward_redemption_status(
                token, self._broadcaster_id, self.id, reward_id, False
            )
        except Unauthorized as error:
            raise Unauthorized("The given token is invalid", "", 401) from error
        except HTTPException as error:
            status = error.args[2]
            if status == 403:
                raise HTTPException(
                    "The custom reward was created by a different application, or channel points are "
                    "not available for the broadcaster (403)",
                    error.args[1],
                    403,
                ) from error
            raise
        else:
            self.__init__(data[0], self._http, self.reward if isinstance(self.reward, CustomReward) else None)
            return self



"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import asyncio
import copy
import datetime
import logging
from typing import TYPE_CHECKING, Union, List, Tuple, Any, Dict, Optional
from typing_extensions import Literal

import aiohttp
from yarl import URL

from . import errors
from .cooldowns import RateBucket

try:
    import ujson as json
except Exception:
    import json
if TYPE_CHECKING:
    from .client import Client
logger = logging.getLogger("twitchio.http")


class Route:
    BASE_URL = "https://api.twitch.tv/helix"

    __slots__ = "path", "body", "headers", "query", "method"

    def __init__(
        self,
        method: str,
        path: Union[str, URL],
        body: Union[str, dict] = None,
        query: List[Tuple[str, Any]] = None,
        headers: dict = None,
        token: str = None,
    ):
        self.headers = headers or {}
        self.method = method
        self.query = query

        if token:
            self.headers["Authorization"] = "Bearer " + token
        if isinstance(path, URL):
            self.path = path
        else:
            self.path = URL(self.BASE_URL + "/" + path.rstrip("/"))
        if query:
            self.path = self.path.with_query(query)
        if isinstance(body, dict):
            self.body = json.dumps(body)
            self.headers["Content-Type"] = "application/json"
        else:
            self.body = body


class TwitchHTTP:
    TOKEN_BASE = "https://id.twitch.tv/oauth2/token"

    def __init__(
        self, client: "Client", *, api_token: str = None, client_secret: str = None, client_id: str = None, **kwargs
    ):
        self.client = client
        self.session = None
        self.token = api_token
        self.app_token = None
        self._refresh_token = None
        self.client_secret = client_secret
        self.client_id = client_id
        self.nick = None
        self.user_id: Optional[int] = None

        self.bucket = RateBucket(method="http")
        self.scopes = kwargs.get("scopes", [])

    async def request(self, route: Route, *, paginate=True, limit=100, full_body=False, force_app_token=False):
        """
        Fulfills an API request

        Parameters
        -----------
        route : :class:`twitchio.http.Route`
            The route to follow
        paginate : :class:`bool`
            whether or not to paginate the requests where possible. Defaults to True
        limit : :class:`int`
            The data limit per request when paginating. Defaults to 100
        full_body : class:`bool`
            Whether to return the full response body or to accumulate the `data` key. Defaults to False. `paginate` must be False if this is True.
        force_app_token : :class:`bool`
            Forcibly use the client_id and client_secret generated token, if available. Otherwise fail the request immediately
        """
        if full_body:
            assert not paginate
        if (not self.client_id or not self.nick) and self.token:
            await self.validate(token=self.token)
        if not self.client_id:
            raise errors.NoClientID("A Client ID is required to use the Twitch API")
        headers = route.headers or {}

        await self._apply_auth(headers, force_app_token, False)

        headers["Client-ID"] = self.client_id

        if not self.session:
            self.session = aiohttp.ClientSession()
        if self.bucket.limited:
            await self.bucket
        cursor = None
        data = []

        def reached_limit():
            return limit and len(data) >= limit

        def get_limit():
            if limit is None:
                return "100"
            to_get = limit - len(data)
            return str(to_get) if to_get < 100 else "100"

        is_finished = False
        while not is_finished:
            path = copy.copy(route.path)

            if limit is not None and paginate:
                q = route.query or []
                if cursor is not None:
                    q = [("after", cursor), *q]
                q = [("first", get_limit()), *q]
                path = path.with_query(q)
            body, is_text = await self._request(route, path, headers, force_app_token=force_app_token)
            if is_text:
                return body
            if full_body:
                return body
            data += body["data"]

            try:
                cursor = body["pagination"].get("cursor", None)
            except KeyError:
                break
            else:
                if not cursor:
                    break
            is_finished = reached_limit() if limit is not None else True if paginate else True
        return data

    async def _apply_auth(self, headers: dict, force_app_token: bool, force_apply: bool) -> None:
        if force_app_token and "Authorization" not in headers:
            if not self.client_secret:
                raise errors.NoToken(
                    "An app access token is required for this route, please provide a client id and client secret"
                )
            if self.app_token is None:
                await self._generate_login()
                headers["Authorization"] = f"Bearer {self.app_token}"
        elif not self.token and not self.client_secret and "Authorization" not in headers:
            raise errors.NoToken(
                "Authorization is required to use the Twitch API. Pass token and/or client_secret to the Client constructor"
            )
        if "Authorization" not in headers or force_apply:
            if not self.token:
                await self._generate_login()

            headers["Authorization"] = f"Bearer {self.token}"

    async def _request(self, route, path, headers, utilize_bucket=True, force_app_token: bool = False):
        reason = None

        for attempt in range(5):
            if utilize_bucket and self.bucket.limited:
                await self.bucket.wait_reset()
            async with self.session.request(route.method, path, headers=headers, data=route.body) as resp:
                try:
                    message = await resp.text(encoding="utf-8")
                    logger.debug(f"Received a response from a request with status {resp.status}: {message}")
                except Exception:
                    message = None
                    logger.debug(f"Received a response from a request with status {resp.status} and without body")

                if 500 <= resp.status <= 504:
                    reason = resp.reason
                    await asyncio.sleep(2**attempt + 1)
                    continue

                if utilize_bucket:
                    reset = resp.headers.get("Ratelimit-Reset")
                    remaining = resp.headers.get("Ratelimit-Remaining")

                    self.bucket.update(reset=reset, remaining=remaining)

                if 200 <= resp.status < 300:
                    if resp.content_type == "application/json" and message:
                        return json.loads(message), False

                    return message, True

                elif resp.status == 400:
                    message_json = json.loads(message)
                    raise errors.HTTPException(
                        f"Failed to fulfill the request", reason=message_json.get("message", ""), status=resp.status
                    )

                elif resp.status == 401:
                    message_json = json.loads(message)
                    if "Invalid OAuth token" in message_json.get("message", ""):
                        try:
                            await self._generate_login()
                            await self._apply_auth(headers, force_app_token, True)
                            continue
                        except:
                            raise errors.Unauthorized(
                                "Your oauth token is invalid, and a new one could not be generated"
                            )

                    raise errors.Unauthorized(
                        "You're not authorized to use this route.",
                        reason=message_json.get("message", ""),
                        status=resp.status,
                    )

                elif resp.status == 429:
                    reason = "Ratelimit Reached"

                    if not utilize_bucket:  # non Helix APIs don't have ratelimit headers
                        await asyncio.sleep(3**attempt + 1)

                    continue

                raise errors.HTTPException(f"Failed to fulfill the request", reason=resp.reason, status=resp.status)
        raise errors.HTTPException("Failed to reach Twitch API", reason=reason, status=resp.status)

    async def _generate_login(self):
        try:
            token = await self.client.event_token_expired()
            if token is not None:
                assert isinstance(token, str), TypeError(f"Expected a string, got {type(token)}")
                self.token = self.app_token = token
                return
        except Exception as e:
            self.client.run_event("error", e)
        if not self.client_id or not self.client_secret:
            raise errors.HTTPException("Unable to generate a token, client id and/or client secret not given")
        if self._refresh_token:
            url = (
                self.TOKEN_BASE
                + "?grant_type=refresh_token&refresh_token={0}&client_id={1}&client_secret={2}".format(
                    self._refresh_token, self.client_id, self.client_secret
                )
            )
        else:
            url = self.TOKEN_BASE + "?client_id={0}&client_secret={1}&grant_type=client_credentials".format(
                self.client_id, self.client_secret
            )
            if self.scopes:
                url += "&scope=" + " ".join(self.scopes)
        if not self.session:
            self.session = aiohttp.ClientSession()
        async with self.session.post(url) as resp:
            if resp.status > 300 or resp.status < 200:
                raise errors.HTTPException("Unable to generate a token: " + await resp.text())
            data = await resp.json()
            self.token = self.app_token = data["access_token"]
            self._refresh_token = data.get("refresh_token", None)
            logger.info("Invalid or no token found, generated new token: %s", self.token)

    async def validate(self, *, token: str = None) -> dict:
        if not token:
            token = self.token
        if not self.session:
            self.session = aiohttp.ClientSession()
        url = "https://id.twitch.tv/oauth2/validate"
        headers = {"Authorization": f"OAuth {token}"}

        async with self.session.get(url, headers=headers) as resp:
            if resp.status == 401:
                raise errors.AuthenticationError("Invalid or unauthorized Access Token passed.")
            if resp.status > 300 or resp.status < 200:
                raise errors.HTTPException("Unable to validate Access Token: " + await resp.text())
            data: dict = await resp.json()
        if not self.nick:
            self.nick = data.get("login")
            self.user_id = data.get("user_id") and int(data["user_id"])
            self.client_id = data.get("client_id")
        return data

    async def post_commercial(self, token: str, broadcaster_id: str, length: int):
        assert length in {30, 60, 90, 120, 150, 180}
        data = await self.request(
            Route(
                "POST", "channels/commercial", body={"broadcaster_id": broadcaster_id, "length": length}, token=token
            ),
            paginate=False,
        )
        data = data[0]
        if data["message"]:
            raise errors.HTTPException(data["message"], extra=data["retry_after"])

    async def get_extension_analytics(
        self,
        token: str,
        extension_id: str = None,
        type: str = None,
        started_at: datetime.datetime = None,
        ended_at: datetime.datetime = None,
    ):
        raise NotImplementedError  # TODO

    async def get_game_analytics(
        self,
        token: str,
        game_id: str = None,
        type: str = None,
        started_at: datetime.datetime = None,
        ended_at: datetime.datetime = None,
    ):
        raise NotImplementedError  # TODO

    async def get_bits_board(
        self,
        token: str,
        period: str = "all",
        user_id: Optional[str] = None,
        started_at: Optional[datetime.datetime] = None,
    ):
        assert period in {"all", "day", "week", "month", "year"}
        query = [
            ("period", period),
            ("started_at", started_at.isoformat() if started_at else None),
            ("user_id", user_id),
        ]

        route = Route(
            "GET",
            "bits/leaderboard",
            "",
            query=[q for q in query if q[1] is not None],
            token=token,
        )
        return await self.request(route, full_body=True, paginate=False)

    async def get_cheermotes(self, broadcaster_id: str):
        return await self.request(Route("GET", "bits/cheermotes", "", query=[("broadcaster_id", broadcaster_id)]))

    async def get_channel_emotes(self, broadcaster_id: str):
        return await self.request(Route("GET", "chat/emotes", "", query=[("broadcaster_id", broadcaster_id)]))

    async def get_global_emotes(self):
        return await self.request(Route("GET", "chat/emotes/global", ""))

    async def get_extension_transactions(self, extension_id: str, ids: List[Any] = None):
        q = [("extension_id", extension_id)]
        if ids:
            q.extend(("id", id) for id in ids)
        return await self.request(Route("GET", "extensions/transactions", "", query=q))

    async def create_reward(
        self,
        token: str,
        broadcaster_id: int,
        title: str,
        cost: int,
        prompt: Optional[str] = None,
        is_enabled: Optional[bool] = True,
        background_color: Optional[str] = None,
        user_input_required: Optional[bool] = False,
        max_per_stream: Optional[int] = None,
        max_per_user: Optional[int] = None,
        global_cooldown: Optional[int] = None,
        fufill_immediatly: Optional[bool] = False,
    ):
        params = [("broadcaster_id", str(broadcaster_id))]
        data = {
            "title": title,
            "cost": cost,
            "prompt": prompt,
            "is_enabled": is_enabled,
            "is_user_input_required": user_input_required,
            "should_redemptions_skip_request_queue": fufill_immediatly,
        }
        if max_per_stream:
            data["max_per_stream"] = max_per_stream
            data["is_max_per_stream_enabled"] = True
        if max_per_user:
            data["max_per_user_per_stream"] = max_per_user
            data["is_max_per_user_per_stream_enabled"] = True
        if background_color:
            data["background_color"] = background_color
        if global_cooldown:
            data["global_cooldown_seconds"] = global_cooldown
            data["is_global_cooldown_enabled"] = True
        return await self.request(Route("POST", "channel_points/custom_rewards", query=params, body=data, token=token))

    async def get_rewards(self, token: str, broadcaster_id: int, only_manageable: bool = False, ids: List[int] = None):
        params = [("broadcaster_id", str(broadcaster_id)), ("only_manageable_rewards", str(only_manageable))]

        if ids:
            params.extend(("id", str(id)) for id in ids)
        return await self.request(Route("GET", "channel_points/custom_rewards", query=params, token=token))

    async def update_reward(
        self,
        token: str,
        broadcaster_id: int,
        reward_id: str,
        title: Optional[str] = None,
        prompt: Optional[str] = None,
        cost: Optional[int] = None,
        background_color: Optional[str] = None,
        enabled: Optional[bool] = None,
        input_required: Optional[bool] = None,
        max_per_stream_enabled: Optional[bool] = None,
        max_per_stream: Optional[int] = None,
        max_per_user_per_stream_enabled: Optional[bool] = None,
        max_per_user_per_stream: Optional[int] = None,
        global_cooldown_enabled: Optional[bool] = None,
        global_cooldown: Optional[int] = None,
        paused: Optional[bool] = None,
        redemptions_skip_queue: Optional[bool] = None,
    ):
        data = {
            "title": title,
            "prompt": prompt,
            "cost": cost,
            "background_color": background_color,
            "is_enabled": enabled,
            "is_user_input_required": input_required,
            "is_max_per_stream_enabled": max_per_stream_enabled,
            "max_per_stream": max_per_stream,
            "is_max_per_user_per_stream_enabled": max_per_user_per_stream_enabled,
            "max_per_user_per_stream": max_per_user_per_stream,
            "is_global_cooldown_enabled": global_cooldown_enabled,
            "global_cooldown_seconds": global_cooldown,
            "is_paused": paused,
            "should_redemptions_skip_request_queue": redemptions_skip_queue,
        }

        data = {k: v for k, v in data.items() if v is not None}

        if not data:
            raise ValueError("Nothing changed!")
        params = [("broadcaster_id", str(broadcaster_id)), ("id", str(reward_id))]
        return await self.request(
            Route(
                "PATCH",
                "channel_points/custom_rewards",
                query=params,
                headers={"Authorization": f"Bearer {token}"},
                body=data,
            )
        )

    async def delete_custom_reward(self, token: str, broadcaster_id: int, reward_id: str):
        params = [("broadcaster_id", str(broadcaster_id)), ("id", reward_id)]
        return await self.request(Route("DELETE", "channel_points/custom_rewards", query=params, token=token))

    async def get_reward_redemptions(
        self,
        token: str,
        broadcaster_id: int,
        reward_id: str,
        redemption_id: Optional[str] = None,
        status: Optional[str] = None,
        sort: str = "OLDEST",
        first: int = 20,
    ):
        params = [("broadcaster_id", str(broadcaster_id)), ("reward_id", reward_id), ("first", first)]
        if redemption_id:
            params.append(("id", redemption_id))
        if status:
            params.append(("status", status))
        if sort:
            params.append(("sort", sort))
        return await self.request(Route("GET", "channel_points/custom_rewards/redemptions", query=params, token=token))

    async def update_reward_redemption_status(
        self, token: str, broadcaster_id: int, reward_id: str, custom_reward_id: str, status: bool
    ):
        params = [("id", reward_id), ("broadcaster_id", str(broadcaster_id)), ("reward_id", custom_reward_id)]
        status = "FULFILLED" if status else "CANCELED"
        return await self.request(
            Route(
                "PATCH",
                "channel_points/custom_rewards/redemptions",
                query=params,
                body={"status": status},
                token=token,
            )
        )

    async def get_predictions(
        self,
        token: str,
        broadcaster_id: int,
        prediction_id: Optional[str] = None,
    ):
        params = [("broadcaster_id", str(broadcaster_id))]

        if prediction_id:
            params.extend(("prediction_id", prediction_id))
        return await self.request(Route("GET", "predictions", query=params, token=token), paginate=False)

    async def patch_prediction(
        self, token: str, broadcaster_id: str, prediction_id: str, status: str, winning_outcome_id: Optional[str] = None
    ):
        body = {
            "broadcaster_id": broadcaster_id,
            "id": prediction_id,
            "status": status,
        }

        if status == "RESOLVED" and winning_outcome_id:
            body["winning_outcome_id"] = winning_outcome_id

        return await self.request(
            Route(
                "PATCH",
                "predictions",
                body=body,
                token=token,
            )
        )

    async def post_prediction(
        self, token: str, broadcaster_id: int, title: str, blue_outcome: str, pink_outcome: str, prediction_window: int
    ):
        body = {
            "broadcaster_id": broadcaster_id,
            "title": title,
            "prediction_window": prediction_window,
            "outcomes": [
                {
                    "title": blue_outcome,
                },
                {
                    "title": pink_outcome,
                },
            ],
        }
        return await self.request(
            Route("POST", "predictions", body=body, token=token),
            paginate=False,
        )

    async def post_create_clip(self, token: str, broadcaster_id: int, has_delay=False):
        return await self.request(
            Route(
                "POST", "clips", query=[("broadcaster_id", broadcaster_id), ("has_delay", str(has_delay))], token=token
            ),
            paginate=False,
        )

    async def get_clips(
        self,
        broadcaster_id: int = None,
        game_id: str = None,
        ids: Optional[List[str]] = None,
        started_at: Optional[datetime.datetime] = None,
        ended_at: Optional[datetime.datetime] = None,
        is_featured: Optional[bool] = None,
        token: Optional[str] = None,
    ):
        if started_at and started_at.tzinfo is None:
            started_at = started_at.replace(tzinfo=datetime.timezone.utc)
        if ended_at and ended_at.tzinfo is None:
            ended_at = ended_at.replace(tzinfo=datetime.timezone.utc)

        q = [
            ("broadcaster_id", broadcaster_id),
            ("game_id", game_id),
            ("started_at", started_at.isoformat() if started_at else None),
            ("ended_at", ended_at.isoformat() if ended_at else None),
            ("is_featured", str(is_featured) if is_featured is not None else None),
        ]
        if ids:
            q.extend(("id", id) for id in ids)
        query = [x for x in q if x[1] is not None]

        return await self.request(Route("GET", "clips", query=query, token=token))

    async def post_entitlements_upload(self, manifest_id: str, type="bulk_drops_grant"):
        return await self.request(
            Route("POST", "entitlements/upload", query=[("manifest_id", manifest_id), ("type", type)])
        )

    async def get_entitlements(self, id: str = None, user_id: str = None, game_id: str = None):
        return await self.request(
            Route("GET", "entitlements/drops", query=[("id", id), ("user_id", user_id), ("game_id", game_id)])
        )

    async def get_code_status(self, codes: List[str], user_id: int):
        q = [("user_id", user_id)]
        q.extend(("code", code) for code in codes)

        return await self.request(Route("GET", "entitlements/codes", query=q))

    async def post_redeem_code(self, user_id: int, codes: List[str]):
        q = [("user_id", user_id)]
        q.extend(("code", c) for c in codes)

        return await self.request(Route("POST", "entitlements/code", query=q))

    async def get_top_games(self):
        return await self.request(Route("GET", "games/top"))

    async def get_games(
        self, game_ids: Optional[List[Any]], game_names: Optional[List[str]], igdb_ids: Optional[List[int]]
    ):
        if not any((game_ids, game_names, igdb_ids)):
            raise ValueError("At least one of game id, name or IGDB id must be provided.")
        q = []
        if game_ids:
            q.extend(("id", id) for id in game_ids)
        if game_names:
            q.extend(("name", name) for name in game_names)
        if igdb_ids:
            q.extend(("igdb_id", id) for id in igdb_ids)
        return await self.request(Route("GET", "games", query=q))

    async def get_hype_train(self, broadcaster_id: str, id: Optional[str] = None, token: str = None):
        return await self.request(
            Route(
                "GET",
                "hypetrain/events",
                query=[x for x in [("broadcaster_id", broadcaster_id), ("id", id)] if x[1] is not None],
                token=token,
            )
        )

    async def post_automod_check(self, token: str, broadcaster_id: str, *msgs: List[Dict[str, str]]):
        return await self.request(
            Route(
                "POST",
                "moderation/enforcements/status",
                query=[("broadcaster_id", broadcaster_id)],
                body={"data": msgs},
                token=token,
            )
        )

    async def post_snooze_ad(self, token: str, broadcaster_id: str):
        q = [("broadcaster_id", broadcaster_id)]
        return await self.request(Route("POST", "channels/ads/schedule/snooze", query=q, token=token))

    async def get_ad_schedule(self, token: str, broadcaster_id: str):
        q = [("broadcaster_id", broadcaster_id)]
        return await self.request(Route("GET", "channels/ads", query=q, token=token))

    async def get_channel_ban_unban_events(self, token: str, broadcaster_id: str, user_ids: List[str] = None):
        q = [("broadcaster_id", broadcaster_id)]
        if user_ids:
            q.extend(("user_id", id) for id in user_ids)
        return await self.request(Route("GET", "moderation/banned/events", query=q, token=token))

    async def get_channel_bans(self, token: str, broadcaster_id: str, user_ids: List[str] = None):
        q = [("broadcaster_id", broadcaster_id)]
        if user_ids:
            q.extend(("user_id", id) for id in user_ids)
        return await self.request(Route("GET", "moderation/banned", query=q, token=token))

    async def get_moderated_channels(self, token: str, user_id: str):
        q = [("user_id", user_id)]
        return await self.request(Route("GET", "moderation/channels", query=q, token=token))

    async def get_channel_moderators(self, token: str, broadcaster_id: str, user_ids: List[str] = None):
        q = [("broadcaster_id", broadcaster_id)]
        if user_ids:
            q.extend(("user_id", id) for id in user_ids)
        return await self.request(Route("GET", "moderation/moderators", query=q, token=token))

    async def get_channel_mod_events(self, token: str, broadcaster_id: str, user_ids: List[str] = None):
        q = [("broadcaster_id", broadcaster_id)]
        q.extend(("user_id", id) for id in user_ids)

        return await self.request(Route("GET", "moderation/moderators/events", query=q, token=token))

    async def get_search_categories(self, query: str, token: str = None):
        return await self.request(Route("GET", "search/categories", query=[("query", query)], token=token))

    async def get_search_channels(self, query: str, token: str = None, live: bool = False):
        return await self.request(
            Route("GET", "search/channels", query=[("query", query), ("live_only", str(live))], token=token)
        )

    async def get_user_emotes(self, user_id: str, broadcaster_id: Optional[str], token: str):
        q: List = [("user_id", user_id)]
        if broadcaster_id:
            q.append(("broadcaster_id", broadcaster_id))

        return await self.request(Route("GET", "chat/emotes/user", query=q, token=token))

    async def get_stream_key(self, token: str, broadcaster_id: str):
        return await self.request(
            Route("GET", "streams/key", query=[("broadcaster_id", broadcaster_id)], token=token), paginate=False
        )

    async def get_streams(
        self,
        game_ids: Optional[List[int]] = None,
        user_ids: Optional[List[int]] = None,
        user_logins: Optional[List[str]] = None,
        languages: Optional[List[str]] = None,
        type_: Literal["all", "live"] = "all",
        token: Optional[str] = None,
    ):
        q = [("type", type_)]
        if game_ids:
            q.extend(("game_id", str(g)) for g in game_ids)
        if user_ids:
            q.extend(("user_id", str(u)) for u in user_ids)
        if user_logins:
            q.extend(("user_login", l) for l in user_logins)
        if languages:
            q.extend(("language", l) for l in languages)
        return await self.request(Route("GET", "streams", query=q, token=token))

    async def post_stream_marker(self, token: str, user_id: str, description: str = None):
        return await self.request(
            Route("POST", "streams/markers", body={"user_id": user_id, "description": description}, token=token)
        )

    async def get_stream_markers(self, token: str, user_id: str = None, video_id: str = None):
        return await self.request(
            Route(
                "GET",
                "streams/markers",
                query=[x for x in [("user_id", user_id), ("video_id", video_id)] if x[1] is not None],
                token=token,
            )
        )

    async def get_channels(self, broadcaster_id: str, token: Optional[str] = None):
        return await self.request(Route("GET", "channels", query=[("broadcaster_id", broadcaster_id)], token=token))

    async def get_channels_new(self, broadcaster_ids: List[int], token: Optional[str] = None):
        if len(broadcaster_ids) > 100:
            raise ValueError("Maximum of 100 broadcaster_ids")
        q = [("broadcaster_id", str(broadcaster_id)) for broadcaster_id in broadcaster_ids]
        return await self.request(Route("GET", "channels", query=q, token=token))

    async def patch_channel(
        self,
        token: str,
        broadcaster_id: str,
        game_id: str = None,
        language: str = None,
        title: str = None,
        content_classification_labels: List[Dict[str, Union[str, bool]]] = None,
        is_branded_content: bool = None,
    ):
        assert any((game_id, language, title, content_classification_labels, is_branded_content))
        body = {
            k: v
            for k, v in {
                "game_id": game_id,
                "broadcaster_language": language,
                "title": title,
                "is_branded_content": is_branded_content,
            }.items()
            if v is not None
        }

        if content_classification_labels is not None:
            body["content_classification_labels"] = content_classification_labels

        return await self.request(
            Route("PATCH", "channels", query=[("broadcaster_id", broadcaster_id)], body=body, token=token)
        )

    async def get_channel_schedule(
        self,
        broadcaster_id: str,
        segment_ids: Optional[List[str]] = None,
        start_time: Optional[datetime.datetime] = None,
        utc_offset: Optional[int] = None,
        first: int = 20,
    ):
        if first > 25 or first < 1:
            raise ValueError("The parameter 'first' was malformed: the value must be less than or equal to 25")
        if segment_ids is not None and len(segment_ids) > 100:
            raise ValueError("segment_id can only have 100 entries")
        if start_time:
            start_time = start_time.strftime("%Y-%m-%dT%H:%M:%SZ")
        if utc_offset:
            utc_offset = str(utc_offset)
        q = [
            x
            for x in [
                ("broadcaster_id", broadcaster_id),
                ("first", first),
                ("start_time", start_time),
                ("utc_offset", utc_offset),
            ]
            if x[1] is not None
        ]

        if segment_ids:
            q.extend(("id", id) for id in segment_ids)
        return await self.request(Route("GET", "schedule", query=q), paginate=False, full_body=True)

    async def get_channel_subscriptions(self, token: str, broadcaster_id: str, user_ids: Optional[List[str]] = None):
        q = [("broadcaster_id", broadcaster_id)]
        if user_ids:
            q.extend(("user_id", u) for u in user_ids)
        return await self.request(Route("GET", "subscriptions", query=q, token=token))

    async def get_stream_tags(self, tag_ids: Optional[List[str]] = None):
        q = []
        if tag_ids:
            q.extend(("tag_id", u) for u in tag_ids)
        return await self.request(Route("GET", "tags/streams", query=q or None))

    async def get_channel_tags(self, broadcaster_id: str):
        return await self.request(Route("GET", "streams/tags", query=[("broadcaster_id", broadcaster_id)]))

    async def put_replace_channel_tags(self, token: str, broadcaster_id: str, tag_ids: List[str] = None):
        return await self.request(
            Route(
                "PUT",
                "streams/tags",
                query=[("broadcaster_id", broadcaster_id)],
                body={"tag_ids": tag_ids},
                token=token,
            )
        )

    async def post_follow_channel(self, token: str, from_id: str, to_id: str, notifications=False):
        return await self.request(
            Route(
                "POST",
                "users/follows",
                query=[("from_id", from_id), ("to_id", to_id), ("allow_notifications", str(notifications))],
                token=token,
            )
        )

    async def delete_unfollow_channel(self, token: str, from_id: str, to_id: str):
        return await self.request(
            Route("DELETE", "users/follows", query=[("from_id", from_id), ("to_id", to_id)], token=token)
        )

    async def get_users(self, ids: List[int], logins: List[str], token: Optional[str] = None):
        q = []
        if ids:
            q.extend(("id", id) for id in ids)
        if logins:
            q.extend(("login", login) for login in logins)
        return await self.request(Route("GET", "users", query=q, token=token))

    async def get_user_follows(
        self, from_id: Optional[str] = None, to_id: Optional[str] = None, token: Optional[str] = None
    ):
        return await self.request(
            Route(
                "GET",
                "users/follows",
                query=[x for x in [("from_id", from_id), ("to_id", to_id)] if x[1] is not None],
                token=token,
            )
        )

    async def put_update_user(self, token: str, description: str):
        return await self.request(Route("PUT", "users", query=[("description", description)], token=token))

    async def get_channel_extensions(self, token: str):
        return await self.request(Route("GET", "users/extensions/list", token=token))

    async def get_user_active_extensions(self, token: str, user_id: str = None):
        return (
            await self.request(
                Route("GET", "users/extensions", query=[("user_id", user_id)], token=token),
                paginate=False,
                full_body=True,
            )
        )["data"]

    async def put_user_extensions(self, token: str, data: Dict[str, Any]):
        return (
            await self.request(
                Route("PUT", "users/extensions", token=token, body={"data": data}), paginate=False, full_body=True
            )
        )["data"]

    async def get_videos(
        self,
        ids: List[str] = None,
        user_id: str = None,
        game_id: str = None,
        sort: str = "time",
        type: str = "all",
        period: str = "all",
        language: str = None,
        token: str = None,
    ):
        q = [
            x
            for x in [
                ("user_id", user_id),
                ("game_id", game_id),
                ("sort", sort),
                ("type", type),
                ("period", period),
                ("lanaguage", language),
            ]
            if x[1] is not None
        ]

        if ids:
            q.extend(("id", id) for id in ids)
        return await self.request(Route("GET", "videos", query=q, token=token))

    async def delete_videos(self, token: str, ids: List[int]):
        q = [("id", str(x)) for x in ids]

        return (await self.request(Route("DELETE", "videos", query=q, token=token), paginate=False, full_body=True))[
            "data"
        ]

    async def get_webhook_subs(self):
        return await self.request(Route("GET", "webhooks/subscriptions"))

    async def get_teams(self, team_name: Optional[str] = None, team_id: Optional[str] = None):
        if team_name:
            q = [("name", team_name)]
        elif team_id:
            q = [("id", team_id)]
        else:
            raise ValueError("You need to provide a team name or id")
        return await self.request(Route("GET", "teams", query=q))

    async def get_channel_teams(self, broadcaster_id: str):
        q = [("broadcaster_id", broadcaster_id)]
        return await self.request(Route("GET", "teams/channel", query=q), paginate=False, full_body=True)

    async def get_polls(
        self,
        broadcaster_id: str,
        token: str,
        poll_ids: Optional[List[str]] = None,
        first: Optional[int] = 20,
    ):
        if poll_ids and len(poll_ids) > 100:
            raise ValueError("poll_ids can only have up to 100 entries")
        if first and (first > 25 or first < 1):
            raise ValueError("first can only be between 1 and 20")
        q = [("broadcaster_id", broadcaster_id), ("first", first)]

        if poll_ids:
            q.extend(("id", poll_id) for poll_id in poll_ids)
        return await self.request(Route("GET", "polls", query=q, token=token), paginate=False, full_body=True)

    async def post_poll(
        self,
        broadcaster_id: str,
        token: str,
        title: str,
        choices,
        duration: int,
        bits_voting_enabled: Optional[bool] = False,
        bits_per_vote: Optional[int] = None,
        channel_points_voting_enabled: Optional[bool] = False,
        channel_points_per_vote: Optional[int] = None,
    ):
        if len(title) > 60:
            raise ValueError("title must be less than or equal to 60 characters")
        if len(choices) < 2 or len(choices) > 5:
            raise ValueError("You must have between 2 and 5 choices")
        for c in choices:
            if len(c) > 25:
                raise ValueError("choice title must be less than or equal to 25 characters")
        if duration < 15 or duration > 1800:
            raise ValueError("duration must be between 15 and 1800 seconds")
        if bits_per_vote and bits_per_vote > 10000:
            raise ValueError("bits_per_vote must bebetween 0 and 10000")
        if channel_points_per_vote and channel_points_per_vote > 1000000:
            raise ValueError("channel_points_per_vote must bebetween 0 and 1000000")
        body = {
            "broadcaster_id": broadcaster_id,
            "title": title,
            "choices": [{"title": choice} for choice in choices],
            "duration": duration,
            "bits_voting_enabled": str(bits_voting_enabled),
            "channel_points_voting_enabled": str(channel_points_voting_enabled),
        }
        if bits_voting_enabled and bits_per_vote:
            body["bits_per_vote"] = bits_per_vote
        if channel_points_voting_enabled and channel_points_per_vote:
            body["channel_points_per_vote"] = channel_points_per_vote
        return await self.request(Route("POST", "polls", body=body, token=token))

    async def patch_poll(self, broadcaster_id: str, token: str, id: str, status: str):
        body = {"broadcaster_id": broadcaster_id, "id": id, "status": status}
        return await self.request(Route("PATCH", "polls", body=body, token=token))

    async def get_goals(self, broadcaster_id: str, token: str):
        return await self.request(Route("GET", "goals", query=[("broadcaster_id", broadcaster_id)], token=token))

    async def get_chat_settings(
        self, broadcaster_id: str, token: Optional[str] = None, moderator_id: Optional[str] = None
    ):
        q = [("broadcaster_id", broadcaster_id)]
        if moderator_id and token:
            q.append(("moderator_id", moderator_id))
        return await self.request(Route("GET", "chat/settings", query=q, token=token))

    async def patch_chat_settings(
        self,
        token: str,
        broadcaster_id: str,
        moderator_id: str,
        emote_mode: Optional[bool] = None,
        follower_mode: Optional[bool] = None,
        follower_mode_duration: Optional[int] = None,
        slow_mode: Optional[bool] = None,
        slow_mode_wait_time: Optional[int] = None,
        subscriber_mode: Optional[bool] = None,
        unique_chat_mode: Optional[bool] = None,
        non_moderator_chat_delay: Optional[bool] = None,
        non_moderator_chat_delay_duration: Optional[int] = None,
    ):
        if follower_mode_duration and follower_mode_duration > 129600:
            raise ValueError("follower_mode_duration must be below 129600")
        if slow_mode_wait_time and (slow_mode_wait_time < 3 or slow_mode_wait_time > 120):
            raise ValueError("slow_mode_wait_time must be between 3 and 120")
        if non_moderator_chat_delay_duration and non_moderator_chat_delay_duration not in {2, 4, 6}:
            raise ValueError("non_moderator_chat_delay_duration must be 2, 4 or 6")
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id)]
        data = {
            "emote_mode": emote_mode,
            "follower_mode": follower_mode,
            "follower_mode_duration": follower_mode_duration,
            "slow_mode": slow_mode,
            "slow_mode_wait_time": slow_mode_wait_time,
            "subscriber_mode": subscriber_mode,
            "unique_chat_mode": unique_chat_mode,
            "non_moderator_chat_delay": non_moderator_chat_delay,
            "non_moderator_chat_delay_duration": non_moderator_chat_delay_duration,
        }
        data = {k: v for k, v in data.items() if v is not None}
        return await self.request(Route("PATCH", "chat/settings", query=q, body=data, token=token))

    async def post_chat_announcement(
        self, token: str, broadcaster_id: str, moderator_id: str, message: str, color: Optional[str] = "primary"
    ):
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id)]
        body = {"message": message, "color": color}
        return await self.request(Route("POST", "chat/announcements", query=q, body=body, token=token))

    async def delete_chat_messages(
        self, token: str, broadcaster_id: str, moderator_id: str, message_id: Optional[str] = None
    ):
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id)]
        if message_id:
            q.append(("message_id", message_id))
        return await self.request(Route("DELETE", "moderation/chat", query=q, token=token))

    async def put_user_chat_color(self, token: str, user_id: str, color: str):
        q = [("user_id", user_id), ("color", color)]
        return await self.request(Route("PUT", "chat/color", query=q, token=token))

    async def get_user_chat_color(self, user_ids: List[int], token: Optional[str] = None):
        if len(user_ids) > 100:
            raise ValueError("You can only get up to 100 user chat colors at once")
        q = [("user_id", str(user_id)) for user_id in user_ids]
        return await self.request(Route("GET", "chat/color", query=q, token=token))

    async def post_channel_moderator(self, token: str, broadcaster_id: str, user_id: str):
        q = [("broadcaster_id", broadcaster_id), ("user_id", user_id)]
        return await self.request(Route("POST", "moderation/moderators", query=q, token=token))

    async def delete_channel_moderator(self, token: str, broadcaster_id: str, user_id: str):
        q = [("broadcaster_id", broadcaster_id), ("user_id", user_id)]
        return await self.request(Route("DELETE", "moderation/moderators", query=q, token=token))

    async def get_channel_vips(
        self, token: str, broadcaster_id: str, first: int = 20, user_ids: Optional[List[int]] = None
    ):
        q = [("broadcaster_id", broadcaster_id), ("first", first)]
        if first > 100:
            raise ValueError("You can only get up to 100 VIPs at once")
        if user_ids:
            if len(user_ids) > 100:
                raise ValueError("You can can only specify up to 100 VIPs")
            q.extend(("user_id", str(user_id)) for user_id in user_ids)
        return await self.request(Route("GET", "channels/vips", query=q, token=token))

    async def post_channel_vip(self, token: str, broadcaster_id: str, user_id: str):
        q = [("broadcaster_id", broadcaster_id), ("user_id", user_id)]
        return await self.request(Route("POST", "channels/vips", query=q, token=token))

    async def delete_channel_vip(self, token: str, broadcaster_id: str, user_id: str):
        q = [("broadcaster_id", broadcaster_id), ("user_id", user_id)]
        return await self.request(Route("DELETE", "channels/vips", query=q, token=token))

    async def post_whisper(self, token: str, from_user_id: str, to_user_id: str, message: str):
        q = [("from_user_id", from_user_id), ("to_user_id", to_user_id)]
        body = {"message": message}
        return await self.request(Route("POST", "whispers", query=q, body=body, token=token))

    async def post_raid(self, token: str, from_broadcaster_id: str, to_broadcaster_id: str):
        q = [("from_broadcaster_id", from_broadcaster_id), ("to_broadcaster_id", to_broadcaster_id)]
        return await self.request(Route("POST", "raids", query=q, token=token))

    async def delete_raid(self, token: str, broadcaster_id: str):
        q = [("broadcaster_id", broadcaster_id)]
        return await self.request(Route("DELETE", "raids", query=q, token=token))

    async def post_ban_timeout_user(
        self,
        token: str,
        broadcaster_id: str,
        moderator_id: str,
        user_id: str,
        reason: str,
        duration: Optional[int] = None,
    ):
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id)]
        body = {"data": {"user_id": user_id, "reason": reason}}
        if duration:
            if duration < 1 or duration > 1209600:
                raise ValueError("Duration must be between 1 and 1209600 seconds")
            body["data"]["duration"] = str(duration)
        return await self.request(Route("POST", "moderation/bans", query=q, body=body, token=token))

    async def delete_ban_timeout_user(
        self,
        token: str,
        broadcaster_id: str,
        moderator_id: str,
        user_id: str,
    ):
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id), ("user_id", user_id)]
        return await self.request(Route("DELETE", "moderation/bans", query=q, token=token))

    async def get_follow_count(
        self, from_id: Optional[str] = None, to_id: Optional[str] = None, token: Optional[str] = None
    ):
        return await self.request(
            Route(
                "GET",
                "users/follows",
                query=[x for x in [("from_id", from_id), ("to_id", to_id)] if x[1] is not None],
                token=token,
            ),
            full_body=True,
            paginate=False,
        )

    async def get_shield_mode_status(self, token: str, broadcaster_id: str, moderator_id: str):
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id)]
        return await self.request(
            Route("GET", "moderation/shield_mode", query=q, token=token), paginate=False, full_body=False
        )

    async def put_shield_mode_status(self, token: str, broadcaster_id: str, moderator_id: str, is_active: bool):
        q = [("broadcaster_id", broadcaster_id), ("moderator_id", moderator_id)]
        body = {"is_active": is_active}
        return await self.request(Route("PUT", "moderation/shield_mode", query=q, body=body, token=token))

    async def get_followed_streams(self, broadcaster_id: str, token: str):
        return await self.request(Route("GET", "streams/followed", query=[("user_id", broadcaster_id)], token=token))

    async def post_shoutout(self, token: str, broadcaster_id: str, moderator_id: str, to_broadcaster_id: str):
        q = [
            ("from_broadcaster_id", broadcaster_id),
            ("moderator_id", moderator_id),
            ("to_broadcaster_id", to_broadcaster_id),
        ]
        return await self.request(Route("POST", "chat/shoutouts", query=q, token=token))

    async def get_global_chat_badges(self):
        return await self.request(Route("GET", "chat/badges/global", ""))

    async def get_channel_chat_badges(self, broadcaster_id: str):
        return await self.request(Route("GET", "chat/badges", "", query=[("broadcaster_id", broadcaster_id)]))

    async def get_content_classification_labels(self, locale: str):
        return await self.request(Route("GET", "content_classification_labels", "", query=[("locale", locale)]))

    async def get_channel_charity_campaigns(self, broadcaster_id: str, token: str):
        return await self.request(
            Route("GET", "charity/campaigns", query=[("broadcaster_id", broadcaster_id)], token=token)
        )

    async def get_channel_followers(
        self,
        token: str,
        broadcaster_id: str,
        user_id: Optional[int] = None,
    ):
        query = [("broadcaster_id", broadcaster_id)]

        if user_id is not None:
            query.append(("user_id", str(user_id)))

        return await self.request(
            Route(
                "GET",
                "channels/followers",
                query=query,
                token=token,
            )
        )

    async def get_channel_followed(
        self,
        token: str,
        user_id: str,
        broadcaster_id: Optional[int] = None,
    ):
        query = [("user_id", user_id)]

        if broadcaster_id is not None:
            query.append(("broadcaster_id", str(broadcaster_id)))

        return await self.request(
            Route(
                "GET",
                "channels/followed",
                query=query,
                token=token,
            )
        )

    async def get_channel_follower_count(self, broadcaster_id: str, token: Optional[str] = None):
        return await self.request(
            Route(
                "GET",
                "channels/followers",
                query=[("broadcaster_id", broadcaster_id)],
                token=token,
            ),
            full_body=True,
            paginate=False,
        )

    async def get_channel_followed_count(self, token: str, user_id: str):
        return await self.request(
            Route(
                "GET",
                "channels/followed",
                query=[("user_id", user_id)],
                token=token,
            ),
            full_body=True,
            paginate=False,
        )



"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import datetime
from typing import TYPE_CHECKING, Union, Optional

if TYPE_CHECKING:
    from .channel import Channel
    from .chatter import Chatter, PartialChatter


class HypeChatData:
    """
    Represents information about hype chats.

    Attributes
    -----------
    amount: :class:`int`
        The amount paid.
    exponent: :class:`int`
        Indicates how many decimal points this currency represents partial amounts in.
        Decimal points start from the right side of the value defined in pinned-chat-paid-amount.
    currency: :class:`str`
        The currency paid in, uses ISO 4217 alphabetic currency code (eg. USD, EUR, GBP).
        The ISO 4217 alphabetic currency code the user has sent the Hype Chat in.
    is_system_message: :class:`bool`
        A Boolean value that determines if the message sent with the Hype Chat was filled in by the system.
        If True, the user entered no message and the body message was automatically filled in by the system.
        If False, the user provided their own message to send with the Hype Chat.
    level: :class:`str`
        The level of the Hype Chat, in English. e.g. ONE, TWO, THREE up to TEN.
    """

    def __init__(self, tags: dict) -> None:
        self.amount: int = int(tags["pinned-chat-paid-amount"])
        self.exponent: int = int(tags["pinned-chat-paid-exponent"])
        self.currency: str = tags["pinned-chat-paid-currency"]
        self.is_system_message: bool = tags["pinned-chat-paid-is-system-message"] == "1"
        self.level: str = tags["pinned-chat-paid-level"]


class Message:
    """
    Attributes
    -----------
    content: :class:`str`
        The content of this message.
    echo: :class:`bool`
        Boolean representing if this is a self-message or not.
    first: :class:`bool`
        Boolean representing whether it's a first message or not.
    hype_chat_data: Optional[:class:`HypeChatData`]
        Any hype chat information associated with the message. This will be ``None`` when the message is not a hype chat.
    """

    __slots__ = (
        "_raw_data",
        "content",
        "hype_chat_data",
        "_author",
        "echo",
        "first",
        "_timestamp",
        "_channel",
        "_tags",
        "_id",
    )

    def __init__(self, **kwargs):
        self._raw_data = kwargs.get("raw_data")
        self.content = kwargs.get("content")
        self._author = kwargs.get("author")
        self._channel = kwargs.get("channel")
        self._tags = kwargs.get("tags")
        self.echo = kwargs.get("echo", False)

        self.first = False
        if self._tags is not None:
            first = self._tags.get("first-msg")
            if first == "1":
                self.first = True

        try:
            self._id = self._tags["id"]
            self._timestamp = self._tags["tmi-sent-ts"]
        except KeyError:
            self._id = None
            self._timestamp = datetime.datetime.now().timestamp() * 1000

        if "pinned-chat-paid-amount" in self._tags:
            self.hype_chat_data: Optional[HypeChatData] = HypeChatData(self._tags)
        else:
            self.hype_chat_data: Optional[HypeChatData] = None

    @property
    def id(self) -> str:
        """The Message ID."""
        return self._id

    @property
    def author(self) -> Union["Chatter", "PartialChatter"]:
        """The User object associated with the Message."""
        return self._author

    @property
    def channel(self) -> "Channel":
        """The Channel object associated with the Message."""
        return self._channel

    @property
    def raw_data(self) -> str:
        """The raw data received from Twitch for this Message."""
        return self._raw_data

    @property
    def tags(self) -> dict:
        """The tags associated with the Message.

        Could be None.
        """
        return self._tags

    @property
    def timestamp(self) -> datetime.datetime:
        """The Twitch timestamp for this Message.

        Returns
        ---------
        timestamp:
            UTC datetime object of the Twitch timestamp.
        """
        return datetime.datetime.utcfromtimestamp(int(self._timestamp) / 1000)



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
import asyncio
import datetime
import sys
import traceback
from typing import Callable, Optional


__all__ = ("Routine", "routine")


def compute_timedelta(dt: datetime.datetime) -> float:
    if dt.tzinfo is None:
        dt = dt.astimezone()

    now = datetime.datetime.now(datetime.timezone.utc)
    return max((dt - now).total_seconds(), 0)


class Routine:
    """The main routine class which helps run async background tasks on a schedule.

    Examples
    --------
    .. code:: py

        @routine(seconds=5, iterations=3)
        async def test(arg):
            print(f'Hello {arg}')

        test.start('World!')


    .. warning::

        This class should not be instantiated manually. Use the decorator :func:`routine` instead.
    """

    def __init__(
        self,
        *,
        coro: Callable,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        iterations: Optional[int] = None,
        time: Optional[datetime.datetime] = None,
        delta: Optional[float] = None,
        wait_first: Optional[bool] = False,
    ):
        self._coro = coro
        self._loop = loop or asyncio.get_event_loop()
        self._task: asyncio.Task = None  # type: ignore

        self._time = time
        self._delta = delta

        self._start_time: datetime.datetime = None  # type: ignore

        self._completed_loops = 0

        iterations = iterations if iterations != 0 else None
        self._iterations = iterations
        self._remaining_iterations = iterations

        self._before = None
        self._after = None
        self._error = None

        self._stop_set = False
        self._restarting = False
        self._wait_first = wait_first

        self._stop_on_error = True

        self._instance = None

        self._args: tuple | None = None
        self._kwargs: dict | None = None

    def __get__(self, instance, owner):
        if instance is None:
            return self

        copy = Routine(
            coro=self._coro,
            loop=self._loop,
            iterations=self._iterations,
            time=self._time,
            delta=self._delta,
            wait_first=self._wait_first,
        )

        copy._instance = instance
        copy._before = self._before
        copy._after = self._after
        copy._error = self._error
        setattr(instance, self._coro.__name__, copy)

        return copy

    def start(self, *args, **kwargs) -> asyncio.Task:
        """Start the routine and return the created task.

        Parameters
        ----------
        stop_on_error: Optional[bool]
            Whether or not to stop and cancel the routine on error. Defaults to True.
        \*args
            The args to pass to the routine.
        \*\*kwargs
            The kwargs to pass to the routine.

        Returns
        -------
        :class:`asyncio.Task`
            The created internal asyncio task.

        Raises
        ------
        RuntimeError
            Raised when this routine is already running when start is called.
        """
        if self._task is not None and not self._task.done() and not self._restarting:
            raise RuntimeError(f"Routine {self._coro.__name__!r} is already running and is not done.")

        self._args, self._kwargs = args, kwargs

        self._restarting = False
        self._task = self._loop.create_task(self._routine(*args, **kwargs))

        if not self._error:
            self._error = self.on_error

        return self._task

    def stop(self) -> None:
        """Stop the routine gracefully.

        .. note::

            This allows the current iteration to complete before the routine is cancelled.
            If immediate cancellation is desired consider using :meth:`cancel` instead.
        """
        self._stop_set = True

    def cancel(self) -> None:
        """Cancel the routine effective immediately and non-gracefully.

        .. note::

            Consider using :meth:`stop` if a graceful stop, which will complete the current iteration, is desired.
        """
        if self._can_be_cancelled():
            self._task.cancel()

        if not self._restarting:
            self._task = None

    def restart(self, *args, **kwargs) -> None:
        """Restart the currently running routine.

        Parameters
        ----------
        stop_on_error: Optional[bool]
            Whether or not to stop and cancel the routine on error. Defaults to True.
        force: Optional[bool]
            If True the restart will cancel the currently running routine effective immediately and restart.
            If False a graceful stop will occur, which allows the routine to finish it's current iteration.
            Defaults to True.
        \*args
            The args to pass to the routine.
        \*\*kwargs
            The kwargs to pass to the routine.


        .. note::

            This does not return the internal task unlike :meth:`start`.
        """
        force = kwargs.pop("force", True)
        self._restarting = True

        self._remaining_iterations = self._iterations

        def restart_when_over(fut, *, args=args, kwargs=kwargs):
            self._task.remove_done_callback(restart_when_over)
            self.start(*args, **kwargs)

        if self._can_be_cancelled():
            self._task.add_done_callback(restart_when_over)

            if force:
                self._task.cancel()
            else:
                self.stop()

    def before_routine(self, coro: Callable) -> None:
        """A decorator to assign a coroutine to run before the routine starts."""
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError(f"Expected coroutine function not type, {type(coro).__name__!r}.")

        self._before = coro

    def after_routine(self, coro: Callable) -> None:
        """A decorator to assign a coroutine to run after the routine has ended."""
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError(f"Expected coroutine function not type, {type(coro).__name__!r}.")

        self._after = coro

    def change_interval(
        self,
        *,
        seconds: Optional[float] = 0,
        minutes: Optional[float] = 0,
        hours: Optional[float] = 0,
        time: Optional[datetime.datetime] = None,
        wait_first: Optional[bool] = False,
    ) -> None:
        """Method which schedules the running interval of the task to change.

        Parameters
        ----------
        seconds: Optional[float]
            The seconds to wait before the next iteration of the routine.
        minutes: Optional[float]
            The minutes to wait before the next iteration of the routine.
        hours: Optional[float]
            The hours to wait before the next iteration of the routine.
        time: Optional[datetime.datetime]
            A specific time to run this routine at. If a naive datetime is passed, your system local time will be used.
        wait_first: Optional[bool]
            Whether to wait the specified time before running the first iteration at the new interval.
            Defaults to False.

        Raises
        ------
        RuntimeError
            Raised when the time argument and any hours, minutes or seconds is passed together.


        .. warning::

            The time argument can not be passed in conjunction with hours, minutes or seconds.
            This behaviour is intended as it allows the time to be exact every day.
        """
        time_ = time

        if any((seconds, minutes, hours)) and time_:
            raise RuntimeError(
                "Argument <time> can not be used in conjunction with any <seconds>, <minutes> or <hours> argument(s)."
            )

        if not time_:
            delta = compute_timedelta(
                datetime.datetime.now(datetime.timezone.utc)
                + datetime.timedelta(seconds=seconds, minutes=minutes, hours=hours)
            )

            self._delta = delta
        else:
            now = datetime.datetime.now(time_.tzinfo)
            if time_ < now:
                time_ = datetime.datetime.combine(now.date(), time_.time())
            if time_ < now:
                time_ = time_ + datetime.timedelta(days=1)

            self._time = time_

        kwargs: dict = self._kwargs
        kwargs["force"] = not wait_first

        self.restart(*self._args, **kwargs)

    def error(self, coro: Callable):
        """A decorator to assign a coroutine as the error handler for this routine.

        The error handler takes in one argument: the exception caught.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError(f"Expected coroutine function not type, {type(coro).__name__!r}.")

        self._error = coro

    async def on_error(self, error: Exception):
        """The default error handler for this routine. Can be overwritten with :meth:`error`."""
        print(f"Exception in routine {self._coro.__name__!r}:", file=sys.stderr)
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    @property
    def completed_iterations(self) -> int:
        """A count of completed iterations."""
        return self._completed_loops

    @property
    def remaining_iterations(self) -> Optional[int]:
        """A count of remaining iterations."""
        return self._remaining_iterations

    @property
    def start_time(self) -> Optional[datetime.datetime]:
        """The time the routine was started.

        .. note::

            This does not reset when restarting, stopping or cancelling the routine.
        """
        return self._start_time

    def _can_be_cancelled(self) -> bool:
        return self._task and not self._task.done()

    async def _routine(self, *args, **kwargs) -> None:
        self._stop_on_error = kwargs.pop("stop_on_error", self._stop_on_error)

        self._start_time = datetime.datetime.now(datetime.timezone.utc)

        try:
            if self._before:
                if self._instance:
                    await self._before(self._instance)
                else:
                    await self._before()
        except Exception as e:
            await self._error(e)

            if self._stop_on_error:
                return self.cancel()

        if self._time:
            wait = compute_timedelta(self._time)
            await asyncio.sleep(wait)

        if self._wait_first and not self._time:
            await asyncio.sleep(self._delta)

        if self._remaining_iterations == 0:
            self._remaining_iterations = self._iterations

        iteration: int = 0
        while True:
            iteration += 1
            start = datetime.datetime.now(datetime.timezone.utc)

            try:
                if self._instance:
                    await self._coro(self._instance, *args, **kwargs)
                else:
                    await self._coro(*args, **kwargs)
            except Exception as e:
                await self._error(e)

                if self._stop_on_error:
                    return self.cancel()

            try:
                self._remaining_iterations -= 1
            except TypeError:
                pass
            else:
                if self._remaining_iterations == 0:
                    break

            if self._stop_set:
                self._stop_set = False
                break

            if self._time:
                sleep = compute_timedelta(self._time + datetime.timedelta(days=iteration))
            else:
                sleep = max((start - datetime.datetime.now(datetime.timezone.utc)).total_seconds() + self._delta, 0)

            self._completed_loops += 1
            await asyncio.sleep(sleep)

        try:
            if self._after:
                if self._instance:
                    await self._after(self._instance)
                else:
                    await self._after()
        except Exception as e:
            await self._error(e)
        finally:
            return self.cancel()


def routine(
    *,
    seconds: Optional[float] = 0,
    minutes: Optional[float] = 0,
    hours: Optional[float] = 0,
    time: Optional[datetime.datetime] = None,
    iterations: Optional[int] = None,
    wait_first: Optional[bool] = False,
):
    """A decorator to assign a coroutine as a :class:`Routine`.

    Parameters
    ----------
    seconds: Optional[float]
        The seconds to wait before the next iteration of the routine.
    minutes: Optional[float]
        The minutes to wait before the next iteration of the routine.
    hours: Optional[float]
        The hours to wait before the next iteration of the routine.
    time: Optional[datetime.datetime]
        A specific time to run this routine at. If a naive datetime is passed, your system local time will be used.
    iterations: Optional[int]
        The amount of iterations to run this routine before stopping.
        If set to None or 0, the routine will run indefinitely.
    wait_first: Optional[bool]
        Whether to wait the specified time before running the first iteration.
        This has no effect when the time argument is used. Defaults to False.

    Raises
    ------
    RuntimeError
        Raised when the time argument and any hours, minutes or seconds is passed together.
    TypeError
        Raised when used on a non-coroutine.


    .. warning::

        The time argument can not be passed in conjunction with hours, minutes or seconds.
        This behaviour is intended as it allows the time to be exact every day.
    """

    def decorator(coro: Callable) -> Routine:
        time_ = time

        if any((seconds, minutes, hours)) and time_:
            raise RuntimeError(
                "Argument <time> can not be used in conjunction with any <seconds>, <minutes> or <hours> argument(s)."
            )

        if not time_:
            delta = compute_timedelta(
                datetime.datetime.now(datetime.timezone.utc)
                + datetime.timedelta(seconds=seconds, minutes=minutes, hours=hours)
            )
        else:
            delta = None

            now = datetime.datetime.now(time_.tzinfo)
            if time_ < now:
                time_ = datetime.datetime.combine(now.date(), time_.time())
            if time_ < now:
                time_ = time_ + datetime.timedelta(days=1)

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError(f"Expected coroutine function not type, {type(coro).__name__!r}.")

        return Routine(coro=coro, time=time_, delta=delta, iterations=iterations, wait_first=wait_first)

    return decorator



import asyncio
import logging
import socket
import warnings
from typing import Union, Tuple, Type, Optional, Any
from collections.abc import Iterable

import yarl
from aiohttp import web

from twitchio import Client, PartialUser
from . import models, http

try:
    from ssl import SSLContext
except:
    SSLContext = Any

__all__ = ("EventSubClient",)

logger = logging.getLogger("twitchio.ext.eventsub")

_message_types = {
    "webhook_callback_verification": models.ChallengeEvent,
    "notification": models.NotificationEvent,
    "revocation": models.RevokationEvent,
}


class EventSubClient(web.Application):
    def __init__(self, client: Client, webhook_secret: str, callback_route: str, token: str = None):
        self.client = client
        self.secret = webhook_secret
        self.route = callback_route
        self._http = http.EventSubHTTP(self, token=token)
        super(EventSubClient, self).__init__()
        self.router.add_post(yarl.URL(self.route).path, self._callback)
        self._closing = asyncio.Event()

    async def listen(self, **kwargs):
        self._closing.clear()
        await self.client.loop.create_task(self._run_app(**kwargs))

    def stop(self):
        self._closing.set()

    async def delete_subscription(self, subscription_id: str):
        await self._http.delete_subscription(subscription_id)

    async def delete_all_active_subscriptions(self):
        # A convenience method
        active_subscriptions = await self.get_subscriptions("enabled")
        for subscription in active_subscriptions:
            await self.delete_subscription(subscription.id)

    async def get_subscriptions(
        self, status: Optional[str] = None, sub_type: Optional[str] = None, user_id: Optional[int] = None
    ):
        # All possible statuses are:
        #
        #     enabled: designates that the subscription is in an operable state and is valid.
        #     webhook_callback_verification_pending: webhook is pending verification of the callback specified in the subscription creation request.
        #     webhook_callback_verification_failed: webhook failed verification of the callback specified in the subscription creation request.
        #     notification_failures_exceeded: notification delivery failure rate was too high.
        #     authorization_revoked: authorization for user(s) in the condition was revoked.
        #     user_removed: a user in the condition of the subscription was removed.
        return await self._http.get_subscriptions(status, sub_type, user_id)

    async def subscribe_user_updated(self, user: Union[PartialUser, str, int]):
        if isinstance(user, PartialUser):
            user = user.id

        user = str(user)
        return await self._http.create_webhook_subscription(models.SubscriptionTypes.user_update, {"user_id": user})

    async def subscribe_channel_raid(
        self, from_broadcaster: Union[PartialUser, str, int] = None, to_broadcaster: Union[PartialUser, str, int] = None
    ):
        if (not from_broadcaster and not to_broadcaster) or (from_broadcaster and to_broadcaster):
            raise ValueError("Expected 1 of from_broadcaster or to_broadcaster")

        if from_broadcaster:
            who = "from_broadcaster_user_id"
            broadcaster = from_broadcaster
        else:
            who = "to_broadcaster_user_id"
            broadcaster = to_broadcaster

        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id

        broadcaster = str(broadcaster)
        return await self._http.create_webhook_subscription(models.SubscriptionTypes.raid, {who: broadcaster})

    async def _subscribe_channel_points_reward(
        self, event, broadcaster: Union[PartialUser, str, int], reward_id: str = None
    ):
        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id

        broadcaster = str(broadcaster)
        data = {"broadcaster_user_id": broadcaster}
        if reward_id:
            data["reward_id"] = reward_id

        return await self._http.create_webhook_subscription(event, data)

    async def _subscribe_with_broadcaster(
        self, event: Tuple[str, int, Type[models._DataType]], broadcaster: Union[PartialUser, str, int]
    ):
        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id

        broadcaster = str(broadcaster)
        return await self._http.create_webhook_subscription(event, {"broadcaster_user_id": broadcaster})

    async def _subscribe_with_broadcaster_moderator(
        self,
        event: Tuple[str, int, Type[models._DataType]],
        broadcaster: Union[PartialUser, str, int],
        moderator: Union[PartialUser, str, int],
    ):
        if isinstance(broadcaster, PartialUser):
            broadcaster = broadcaster.id
        if isinstance(moderator, PartialUser):
            moderator = moderator.id

        broadcaster = str(broadcaster)
        moderator = str(moderator)
        return await self._http.create_webhook_subscription(
            event, {"broadcaster_user_id": broadcaster, "moderator_user_id": moderator}
        )

    def subscribe_channel_bans(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.ban, broadcaster)

    def subscribe_channel_unbans(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.unban, broadcaster)

    def subscribe_channel_subscriptions(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription, broadcaster)

    def subscribe_channel_subscription_end(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription_end, broadcaster)

    def subscribe_channel_subscription_gifts(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription_gift, broadcaster)

    def subscribe_channel_subscription_messages(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.subscription_message, broadcaster)

    def subscribe_channel_cheers(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.cheer, broadcaster)

    def subscribe_channel_update(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_update, broadcaster)

    def subscribe_channel_follows(self, broadcaster: Union[PartialUser, str, int]):
        """
        .. warning::
            This endpoint is deprecated, use :func:`~EventSubClient.subscribe_channel_follows_v2`

        """
        warnings.warn(
            "subscribe_channel_follows is deprecated, use subscribe_channel_follows_v2 instead.", DeprecationWarning, 2
        )

        return self._subscribe_with_broadcaster(models.SubscriptionTypes.follow, broadcaster)

    def subscribe_channel_follows_v2(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(models.SubscriptionTypes.followV2, broadcaster, moderator)

    def subscribe_channel_moderators_add(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_moderator_add, broadcaster)

    def subscribe_channel_moderators_remove(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_moderator_remove, broadcaster)

    def subscribe_channel_goal_begin(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_goal_begin, broadcaster)

    def subscribe_channel_goal_progress(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_goal_progress, broadcaster)

    def subscribe_channel_goal_end(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_goal_end, broadcaster)

    def subscribe_channel_hypetrain_begin(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.hypetrain_begin, broadcaster)

    def subscribe_channel_hypetrain_progress(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.hypetrain_progress, broadcaster)

    def subscribe_channel_hypetrain_end(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.hypetrain_end, broadcaster)

    def subscribe_channel_stream_start(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.stream_start, broadcaster)

    def subscribe_channel_stream_end(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.stream_end, broadcaster)

    def subscribe_channel_points_reward_added(self, broadcaster: Union[PartialUser, str, int], reward_id: str):
        return self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_add, broadcaster, reward_id
        )

    def subscribe_channel_points_reward_updated(self, broadcaster: Union[PartialUser, str, int], reward_id: str):
        return self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_update, broadcaster, reward_id
        )

    def subscribe_channel_points_reward_removed(self, broadcaster: Union[PartialUser, str, int], reward_id: str):
        return self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_remove, broadcaster, reward_id
        )

    def subscribe_channel_points_redeemed(self, broadcaster: Union[PartialUser, str, int], reward_id: str = None):
        return self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_redeem, broadcaster, reward_id
        )

    def subscribe_channel_points_redeem_updated(self, broadcaster: Union[PartialUser, str, int], reward_id: str = None):
        return self._subscribe_channel_points_reward(
            models.SubscriptionTypes.channel_reward_redeem_updated, broadcaster, reward_id
        )

    def subscribe_channel_poll_begin(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.poll_begin, broadcaster)

    def subscribe_channel_poll_progress(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.poll_progress, broadcaster)

    def subscribe_channel_poll_end(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.poll_end, broadcaster)

    def subscribe_channel_prediction_begin(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_begin, broadcaster)

    def subscribe_channel_prediction_progress(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_progress, broadcaster)

    def subscribe_channel_prediction_lock(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_lock, broadcaster)

    def subscribe_channel_prediction_end(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.prediction_end, broadcaster)

    def subscribe_channel_auto_reward_redeem(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.auto_reward_redeem, broadcaster)

    def subscribe_channel_shield_mode_begin(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shield_mode_begin, broadcaster, moderator
        )

    def subscribe_channel_shield_mode_end(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shield_mode_end, broadcaster, moderator
        )

    def subscribe_channel_shoutout_create(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shoutout_create, broadcaster, moderator
        )

    def subscribe_channel_shoutout_receive(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_shoutout_receive, broadcaster, moderator
        )

    def subscribe_channel_unban_request_create(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.unban_request_create, broadcaster, moderator
        )

    def subscribe_channel_unban_request_resolve(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.unban_request_resolve, broadcaster, moderator
        )

    def subscribe_automod_message_hold(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_message_hold, broadcaster, moderator
        )

    def subscribe_automod_message_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_message_update, broadcaster, moderator
        )

    def subscribe_automod_settings_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_settings_update, broadcaster, moderator
        )

    def subscribe_automod_terms_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.automod_terms_update, broadcaster, moderator
        )

    def subscribe_channel_charity_donate(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_charity_donate, broadcaster)

    def subscribe_suspicious_user_update(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.suspicious_user_update, broadcaster, moderator
        )

    def subscribe_channel_moderate(
        self, broadcaster: Union[PartialUser, str, int], moderator: Union[PartialUser, str, int]
    ):
        return self._subscribe_with_broadcaster_moderator(
            models.SubscriptionTypes.channel_moderate, broadcaster, moderator
        )

    def subscribe_channel_vip_add(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_vip_add, broadcaster)

    def subscribe_channel_vip_remove(self, broadcaster: Union[PartialUser, str, int]):
        return self._subscribe_with_broadcaster(models.SubscriptionTypes.channel_vip_remove, broadcaster)

    async def subscribe_user_authorization_granted(self):
        return await self._http.create_webhook_subscription(
            models.SubscriptionTypes.user_authorization_grant, {"client_id": self.client._http.client_id}
        )

    async def subscribe_user_authorization_revoked(self):
        return await self._http.create_webhook_subscription(
            models.SubscriptionTypes.user_authorization_revoke, {"client_id": self.client._http.client_id}
        )

    async def _callback(self, request: web.Request) -> web.Response:
        payload = await request.text()
        typ = request.headers.get("Twitch-Eventsub-Message-Type", "")
        if not typ:
            return web.Response(status=404)

        if typ not in _message_types:
            logger.warning(f"Unexpected message type: {typ}")
            return web.Response(status=400)

        logger.debug(f"Recived a message type: {typ}")
        event = _message_types[typ](self, payload, request)
        response = event.verify()

        if response.status != 200:
            return response

        if typ == "notification":
            self.client.run_event(
                f"eventsub_notification_{models.SubscriptionTypes._name_map[event.subscription.type]}", event
            )
        elif typ == "revocation":
            self.client.run_event("eventsub_revokation", event)

        return response

    async def _run_app(
        self,
        *,
        host: Optional[Union[str, web.HostSequence]] = None,
        port: Optional[int] = None,
        path: Optional[str] = None,
        sock: Optional[socket.socket] = None,
        shutdown_timeout: float = 60.0,
        ssl_context: Optional[SSLContext] = None,
        backlog: int = 128,
        access_log_class: Type[web.AbstractAccessLogger] = web.AccessLogger,
        access_log_format: str = web.AccessLogger.LOG_FORMAT,
        access_log: Optional[logging.Logger] = web.access_logger,
        handle_signals: bool = True,
        reuse_address: Optional[bool] = None,
        reuse_port: Optional[bool] = None,
    ) -> None:
        # This function is pulled from aiohttp.web._run_app
        app = self

        runner = web.AppRunner(
            app,
            handle_signals=handle_signals,
            access_log_class=access_log_class,
            access_log_format=access_log_format,
            access_log=access_log,
        )

        await runner.setup()

        sites = []

        try:
            if host is not None:
                if isinstance(host, (str, bytes, bytearray, memoryview)):
                    sites.append(
                        web.TCPSite(
                            runner,
                            host,
                            port,
                            shutdown_timeout=shutdown_timeout,
                            ssl_context=ssl_context,
                            backlog=backlog,
                            reuse_address=reuse_address,
                            reuse_port=reuse_port,
                        )
                    )
                else:
                    for h in host:
                        sites.append(
                            web.TCPSite(
                                runner,
                                h,
                                port,
                                shutdown_timeout=shutdown_timeout,
                                ssl_context=ssl_context,
                                backlog=backlog,
                                reuse_address=reuse_address,
                                reuse_port=reuse_port,
                            )
                        )
            elif path is None and sock is None or port is not None:
                sites.append(
                    web.TCPSite(
                        runner,
                        port=port,
                        shutdown_timeout=shutdown_timeout,
                        ssl_context=ssl_context,
                        backlog=backlog,
                        reuse_address=reuse_address,
                        reuse_port=reuse_port,
                    )
                )

            if path is not None:
                if isinstance(path, (str, bytes, bytearray, memoryview)):
                    sites.append(
                        web.UnixSite(
                            runner,
                            path,
                            shutdown_timeout=shutdown_timeout,
                            ssl_context=ssl_context,
                            backlog=backlog,
                        )
                    )
                else:
                    for p in path:
                        sites.append(
                            web.UnixSite(
                                runner,
                                p,
                                shutdown_timeout=shutdown_timeout,
                                ssl_context=ssl_context,
                                backlog=backlog,
                            )
                        )

            if sock is not None:
                if not isinstance(sock, Iterable):
                    sites.append(
                        web.SockSite(
                            runner,
                            sock,
                            shutdown_timeout=shutdown_timeout,
                            ssl_context=ssl_context,
                            backlog=backlog,
                        )
                    )
                else:
                    for s in sock:
                        sites.append(
                            web.SockSite(
                                runner,
                                s,
                                shutdown_timeout=shutdown_timeout,
                                ssl_context=ssl_context,
                                backlog=backlog,
                            )
                        )
            for site in sites:
                await site.start()

            names = sorted(str(s.name) for s in runner.sites)
            logger.debug("Running EventSub server on {}".format(", ".join(names)))

            await self._closing.wait()
        finally:
            await runner.cleanup()



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
import enum
import time

from .errors import *


__all__ = (
    "Bucket",
    "Cooldown",
)


class Bucket(enum.Enum):
    """
    Enum values for the different cooldown buckets.

    Parameters
    ------------
    default: :class:`enum.Enum`
        The default bucket.
    channel: :class:`enum.Enum`
        Cooldown is shared amongst all chatters per channel.
    member: :class:`enum.Enum`
        Cooldown operates on a per channel basis per user.
    user: :class:`enum.Enum`
        Cooldown operates on a user basis across all channels.
    subscriber: :class:`enum.Enum`
        Cooldown for subscribers.
    mod: :class:`enum.Enum`
        Cooldown for mods.
    """

    default = 0
    channel = 1
    member = 2
    user = 3
    subscriber = 4
    mod = 5


class Cooldown:
    """
    Cooldown decorator values.

    Parameters
    ------------
    rate: :class:`int`
        How many times the command can be invoked before triggering a cooldown inside a time frame.
    per: :class:`float`
        The amount of time in seconds to wait for a cooldown when triggered.
    bucket: :class:`Bucket`
        The bucket that the cooldown is in.

    Examples
    ----------

    .. code:: py

        # Restrict a command to once every 10 seconds on a per channel basis.
        @commands.cooldown(rate=1, per=10, bucket=commands.Bucket.channel)
        @commands.command()
        async def my_command(self, ctx: commands.Context):
            pass

        # Restrict a command to once every 30 seconds for each individual channel a user is in.
        @commands.cooldown(rate=1, per=30, bucket=commands.Bucket.member)
        @commands.command()
        async def my_command(self, ctx: commands.Context):
            pass

        # Restrict a command to 5 times every 60 seconds globally for a user.
        @commands.cooldown(rate=5, per=60, bucket=commands.Bucket.user)
        @commands.command()
        async def my_command(self, ctx: commands.Context):
            pass
    """

    __slots__ = ("_rate", "_per", "bucket", "_window", "_tokens", "_cache")

    def __init__(self, rate: int, per: float, bucket: Bucket):
        self._rate = rate
        self._per = per
        self.bucket = bucket

        self._cache = {}

    def update_bucket(self, ctx):
        now = time.time()

        bucket_keys = self._bucket_keys(ctx)
        buckets = []

        for bucket in bucket_keys:
            (tokens, window) = self._cache[bucket]

            if tokens == self._rate:
                retry = self._per - (now - window)
                raise CommandOnCooldown(command=ctx.command, retry_after=retry)

            tokens += 1

            if tokens == self._rate:
                window = now

            self._cache[bucket] = (tokens, window)

    def reset(self):
        self._cache = {}

    def _bucket_keys(self, ctx):
        buckets = []

        for bucket in ctx.command._cooldowns:
            if bucket.bucket == Bucket.default:
                buckets.append("default")

            if bucket.bucket == Bucket.channel:
                buckets.append(ctx.channel.name)

            if bucket.bucket == Bucket.member:
                buckets.append((ctx.channel.name, ctx.author.id))
            if bucket.bucket == Bucket.user:
                buckets.append(ctx.author.id)

            if bucket.bucket == Bucket.subscriber:
                buckets.append((ctx.channel.name, ctx.author.id, 0))
            if bucket.bucket == Bucket.mod:
                buckets.append((ctx.channel.name, ctx.author.id, 1))

        return buckets

    def _update_cache(self, now=None):
        now = now or time.time()
        dead = [key for key, cooldown in self._cache.items() if now > cooldown[1] + self._per]

        for bucket in dead:
            del self._cache[bucket]

    def get_buckets(self, ctx):
        now = time.time()

        self._update_cache(now)

        bucket_keys = self._bucket_keys(ctx)
        buckets = []

        for index, bucket in enumerate(bucket_keys):
            buckets.append(ctx.command._cooldowns[index])
            if bucket not in self._cache:
                self._cache[bucket] = (0, now)

        return buckets



"""
The MIT License (MIT)

Copyright (c) 2017-present TwitchIO

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

import abc
import time

from .cooldowns import RateBucket
from .errors import *


class IRCLimiterMapping:
    def __init__(self):
        self.buckets = {}

    def get_bucket(self, channel: str, method: str) -> RateBucket:
        try:
            bucket = self.buckets[channel]
        except KeyError:
            bucket = RateBucket(method=method)
            self.buckets[channel] = bucket
        if bucket.method != method:
            bucket.method = method
            bucket.limit = bucket.MODLIMIT if method == "mod" else bucket.IRCLIMIT
            self.buckets[channel] = bucket
        return bucket


limiter = IRCLimiterMapping()


class Messageable(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def _fetch_channel(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _fetch_websocket(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _fetch_message(self):
        raise NotImplementedError

    @abc.abstractmethod
    def _bot_is_mod(self):
        raise NotImplementedError

    def check_bucket(self, channel):
        mod = self._bot_is_mod()

        if mod:
            bucket = limiter.get_bucket(channel=channel, method="mod")
        else:
            bucket = limiter.get_bucket(channel=channel, method="irc")
        now = time.time()
        bucket.update()

        if bucket.limited:
            raise IRCCooldownError(
                f"IRC Message rate limit reached for channel <{channel}>."
                f" Please try again in {bucket._reset - now:.2f}s"
            )

    def check_content(self, content: str):
        if len(content) > 500:
            raise InvalidContent("Content must not exceed 500 characters.")

    async def send(self, content: str):
        """|coro|


        Send a message to the destination associated with the dataclass.

        Destination will either be a channel or user.

        Parameters
        ------------
        content: str
            The content you wish to send as a message. The content must be a string.

        Raises
        --------
        InvalidContent
            Invalid content.
        """
        entity = self._fetch_channel()
        ws = self._fetch_websocket()

        self.check_content(content)
        self.check_bucket(channel=entity.name)

        try:
            name = entity.channel.name
        except AttributeError:
            name = entity.name
        if entity.__messageable_channel__:
            await ws.send(f"PRIVMSG #{name} :{content}\r\n")
        else:
            await ws.send(f"PRIVMSG #jtv :/w {entity.name} {content}\r\n")
